# © Copyright Databand.ai, an IBM Company 2022

import attr

from dbnd._core.parameter.parameter_definition import ParameterDefinition


@attr.s(hash=False, repr=False, str=False)
class _ParameterConfig(ParameterDefinition):
    target = attr.ib(default=None)
    key = attr.ib(default=None)



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging
import random
import time

from dbnd._core.task_build.dbnd_decorator import task
from dbnd._core.tracking.metrics import log_metric


logger = logging.getLogger(__name__)


@task
def dbnd_sanity_check(check_time=datetime.datetime.now(), fail_chance=0.0):
    # type: (datetime.datetime, float) -> str
    logger.info("Running Databand Task Sanity Check!")

    log_metric("metric_check", "OK")

    if random.random() < fail_chance:
        raise Exception("Failed accordingly to fail_chance: %f" % fail_chance)

    log_metric("Happiness Level", "High")

    logger.info("Your system is good to go! Enjoy Databand!")
    return "Databand has been checked at %s" % check_time


@task
def dbnd_simple_task(
    check_time=datetime.datetime.now(), fail_chance=0.0, sleep=0.0, loop=1
):  # type: (datetime.datetime, float, float, int) -> str
    logger.info("Running dbnd_simple_task!")
    log_metric("metric_check", "OK")
    log_metric("metric_random_value", random.random())

    for current in range(loop):
        if current:
            logger.info("Loop %s out of %s", current, loop)
        if sleep:
            logger.info("Sleeping for %s seconds", sleep)
            time.sleep(sleep)

        if random.random() < fail_chance:
            raise Exception("Failed accordingly to fail_chance: %f" % fail_chance)

    return "Databand has been checked at %s" % check_time



# © Copyright Databand.ai, an IBM Company 2022

import logging

from pprint import pprint

import dateutil.parser

from dbnd import get_databand_context
from dbnd._core.errors.base import DatabandApiError


logger = logging.getLogger(__name__)


class DatabandClient(object):
    """
    Customer facing api client
    """

    def __init__(self, api_client, verbose=False):
        self.verbose = verbose
        self.api_client = api_client

    def get_task_run_info(self, task_af_id):
        task_runs_info = self.api_client.api_request(
            "/api/client/v1/task_run_info",
            None,
            method="GET",
            query={
                "filter": '[{"name":"task_af_id","op":"eq","val": "%s"}]' % task_af_id
            },
        )
        task_runs_info = task_runs_info["data"]
        if not task_runs_info:
            raise Exception("Task run not found: %s", task_af_id)
        if len(task_runs_info) > 1:
            logger.warning(
                "More than one task found for %s, returning the last one" % task_af_id
            )

        # return the last one
        # TODO: we need much better query
        return task_runs_info[-1]

    def get_metrics(self, task_run_attempt_uid, source=None):
        query = {"task_run_attempt_uid": task_run_attempt_uid}
        if source:
            query["source"] = source
        metrics = self.api_client.api_request(
            "/api/client/v1/metrics", None, method="GET", query=query
        )

        if self.verbose:
            logger.info(
                "Metrics for %s: %s", task_run_attempt_uid, pprint(metrics["data"])
            )
        return metrics["data"]

    def get_metrics_as_dict(
        self, task_run_attempt_uid, source=None, filter_name_by_prefix=None
    ):
        metrics = self.get_metrics(
            task_run_attempt_uid=task_run_attempt_uid, source=source
        )
        if filter_name_by_prefix:
            # Generator!
            # TODO: move to webserver
            metrics = (m for m in metrics if m["key"].startswith(filter_name_by_prefix))
        return {m["key"]: m for m in metrics}

    def get_run_info(self, run_uid):
        run_info = self.api_client.api_request(
            "/api/client/v1/run/%s" % run_uid, None, method="GET"
        )
        return run_info

    def get_all_runs_info(self):
        runs_info = self.api_client.api_request(
            "/api/client/v1/run_info", None, method="GET"
        )
        return runs_info

    def get_run_info_by_dag_id(self, dag_id):
        runs_info = self.api_client.api_request(
            "/api/client/v1/run_info",
            None,
            method="GET",
            query={"filter": '[{"name":"dag_id","op":"eq","val": "%s"}]' % dag_id},
        )
        return runs_info

    def get_scheduled_jobs_by_scheduled_job_uid(self, scheduled_job_uid):
        scheduled_jobs = self.api_client.api_request(
            "/api/client/v1/scheduled_job_info",
            None,
            method="GET",
            query={
                "filter": '[{"name":"uid","op":"eq","val": "%s"}]' % scheduled_job_uid
            },
        )
        return scheduled_jobs

    def get_scheduled_job_by_job_name(self, dag_id):
        scheduled_job = self.api_client.api_request(
            "/api/client/v1/scheduled_job_info",
            None,
            method="GET",
            query={"filter": '[{"name":"name","op":"eq","val": "%s"}]' % dag_id},
        )
        return scheduled_job

    def get_run_info_by_run_uid(self, run_uid):
        runs_info = self.api_client.api_request(
            "/api/client/v1/run_info",
            None,
            method="GET",
            query={"filter": '[{"name":"uid","op":"eq","val": "%s"}]' % run_uid},
        )
        return runs_info

    def get_airflow_syncers(self):
        af_syncers = self.api_client.api_request(
            "/api/v1/airflow_monitor", None, method="GET"
        )
        return af_syncers

    def archive_syncer(self, tracking_source_uid):
        af_syncers = self.api_client.api_request(
            "/api/v1/airflow_monitor/archive", tracking_source_uid, method="POST"
        )
        return af_syncers

    def unarchive_syncer(self, tracking_source_uid):
        af_syncers = self.api_client.api_request(
            "/api/v1/airflow_monitor/unarchive", tracking_source_uid, method="POST"
        )
        return af_syncers

    def get_airflow_validation(self, syncer_uid):
        validate_response = self.api_client.api_request(
            f"/api/v1/source_monitor/airflow/{syncer_uid}/validation_status",
            None,
            method="GET",
        )
        return validate_response

    def _get_task_start_time(self, task_run):
        if not task_run.get("latest_task_run_attempt"):
            return None
        attempt_end_time = task_run["latest_task_run_attempt"].get("end_date")
        if attempt_end_time is None:
            return None
        return dateutil.parser.isoparse(attempt_end_time)

    def get_first_task_run_error(self, run_uid):
        try:
            run_info = self.get_run_info(run_uid)
        except DatabandApiError as error:
            if error.resp_code == 404:
                return None
            raise

        first_error_time = first_error_task = None
        for task_run in run_info["task_runs"]:
            task_time = self._get_task_start_time(task_run)
            if task_time is None:
                continue
            if first_error_time is None or task_time < first_error_time:
                if task_run["latest_error"] is not None:
                    first_error_time, first_error_task = task_time, task_run

        if first_error_task is None:
            return None

        return first_error_task["latest_error"]

    @classmethod
    def build_databand_client(cls):
        api_client = get_databand_context().databand_api_client
        return DatabandClient(api_client, verbose=True)



# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import logging
import typing

from typing import Any, Dict, Tuple

import attr

from dbnd import ParameterDefinition
from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.constants import (
    RESULT_PARAM,
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    TaskRunState,
)
from dbnd._core.current import (
    current_task_run,
    get_databand_run,
    is_verbose,
    try_get_current_task,
)
from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.log.external_exception_logging import log_exception_to_server
from dbnd._core.parameter.parameter_value import ParameterFilters
from dbnd._core.settings import TrackingConfig
from dbnd._core.task.tracking_task import TrackingTask
from dbnd._core.task_build.task_context import try_get_current_task
from dbnd._core.task_build.task_definition import TaskDefinition
from dbnd._core.task_build.task_results import FuncResultParameter
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.utils.callable_spec import args_to_kwargs
from dbnd._core.utils.timezone import utcnow
from targets import InMemoryTarget, Target
from targets.values import get_value_type_of_obj


if typing.TYPE_CHECKING:
    from dbnd._core.task_build.task_decorator import TaskDecorator

logger = logging.getLogger(__name__)


@attr.s
class TrackedFuncCallWithResult(object):
    call_args = attr.ib()  # type:  Tuple[Any]
    call_kwargs = attr.ib()  # type:  Dict[str,Any]
    callable = attr.ib()
    result = attr.ib(default=None)

    def set_result(self, value):
        self.result = value
        return value

    def invoke(self):
        func = self.callable
        return func(*self.call_args, **self.call_kwargs)


class CallableTrackingManager(object):
    def __init__(self, task_decorator):
        # type: (CallableTrackingManager, TaskDecorator) -> None
        self.task_decorator = task_decorator

        self._tracking_task_definition = None
        self._call_count = 0
        self._call_as_func = False
        self._max_call_count = get_dbnd_project_config().max_calls_per_run

    @property
    def callable(self):
        return self.task_decorator.class_or_func

    def get_tracking_task_definition(self):
        if not self._tracking_task_definition:
            self._tracking_task_definition = self._build_tracking_task_definition()
        return self._tracking_task_definition

    def _build_tracking_task_definition(self):
        return TaskDefinition.from_task_decorator(task_decorator=self.task_decorator)

    def _call_count_limit_exceeded(self):
        if not self._call_as_func:
            self._call_count += 1
            if self._call_count > self._max_call_count:
                logger.info(
                    "Reached maximum tracking limit of {} tasks. Running function regularly.".format(
                        self._max_call_count
                    )
                )
                self._call_as_func = True
        return self._call_as_func

    @contextlib.contextmanager
    def tracking_context(self, call_args, call_kwargs):
        user_code_called = False  # whether we got to executing of user code
        user_code_finished = False  # whether we passed executing of user code
        func_call = None
        try:
            # 1. check that we don't have too many calls
            if self._call_count_limit_exceeded():
                yield _do_nothing_decorator
                return

            # 2. Start or reuse existing "main tracking task" that is root for tracked tasks
            if not try_get_current_task():
                """
                try to get existing task, and if not exists - try to get/create inplace_task_run
                """
                from dbnd._core.tracking.script_tracking_manager import (
                    try_get_inplace_tracking_task_run,
                )

                inplace_tacking_task = try_get_inplace_tracking_task_run()
                if not inplace_tacking_task:
                    # we didn't manage to start inplace tracking task run, we will not be able to track
                    yield _do_nothing_decorator
                    return

            tracking_task_definition = self.get_tracking_task_definition()
            callable_spec = tracking_task_definition.task_decorator.get_callable_spec()

            func_call = TrackedFuncCallWithResult(
                callable=self.callable,
                call_args=tuple(call_args),  # prevent original call_args modification
                call_kwargs=dict(call_kwargs),  # prevent original kwargs modification
            )
            # replace any position argument with kwarg if it possible
            args, kwargs = args_to_kwargs(
                callable_spec.args, func_call.call_args, func_call.call_kwargs
            )

            # instantiate inline task
            task = TrackingTask.for_func(tracking_task_definition, args, kwargs)

            # update upstream/downstream relations - needed for correct tracking
            # we can have the task as upstream , as it was executed already
            parent_task = current_task_run().task
            if not parent_task.task_dag.has_upstream(task):
                parent_task.set_upstream(task)

            # checking if any of the inputs are the outputs of previous task.
            # we can add that task as upstream.
            dbnd_run = get_databand_run()
            call_kwargs_as_targets = dbnd_run.target_origin.get_for_map(kwargs)
            for value_origin in call_kwargs_as_targets.values():
                up_task = value_origin.origin_target.task
                task.set_upstream(up_task)

            # creating task_run as a task we found mid-run
            task_run = dbnd_run.create_task_run_at_execution_time(
                task, task_engine=current_task_run().task_engine
            )

            should_capture_log = (
                TrackingConfig.from_databand_context().capture_tracking_log
            )
            with task_run.runner.task_run_execution_context(
                handle_sigterm=True, capture_log=should_capture_log
            ):
                task_run.set_task_run_state(state=TaskRunState.RUNNING)

                _log_inputs(task_run)

                # if we reached this line, then all tracking initialization is
                # finished successfully, and we're going to execute user code
                user_code_called = True

                try:
                    # tracking_context is context manager - user code will run on yield
                    yield func_call.set_result

                    # if we reached this line, this means that user code finished
                    # successfully without any exceptions
                    user_code_finished = True
                # We catch BaseException since we want to catch KeyboardInterrupts as well
                except BaseException as ex:
                    task_run.finished_time = utcnow()

                    error = TaskRunError.build_from_ex(ex, task_run)
                    task_run.set_task_run_state(TaskRunState.FAILED, error=error)
                    raise

                else:
                    task_run.finished_time = utcnow()

                    # func_call.result should contain result, log it
                    _log_result(task_run, func_call.result)

                    task_run.set_task_run_state(TaskRunState.SUCCESS)
        except BaseException:
            if user_code_called and not user_code_finished:
                # if we started to call the user code and not got to user_code_finished
                # line - it means there was user code exception - so just re-raise it
                raise
            # else it's either we didn't reached calling user code, or already passed it
            # then it's some dbnd tracking error - just log it
            if func_call:
                _handle_tracking_error("tracking-init", func_call)
            else:
                log_exception_to_server()
        # if we didn't reached user_code_called=True line - there was an error during
        # dbnd tracking initialization, so nothing is done - user function wasn't called yet
        if not user_code_called:
            # tracking_context is context manager - user code will run on yield
            yield _do_nothing_decorator
            return


def _handle_tracking_error(msg, func_call=None):
    log_exception_to_server()
    location = " for %s" % func_call.callable if func_call else ""
    msg = "Failed during dbnd %s for %s, ignoring, and continue without tracking" % (
        msg,
        location,
    )
    if is_verbose():
        logger.warning(msg, exc_info=True)
    else:
        logger.info(msg)


def _do_nothing_decorator(f):
    return f


def _log_inputs(task_run):
    """
    For tracking mode. Logs InMemoryTarget inputs.
    """
    try:
        params = task_run.task._params
        for param_value in params.get_param_values(ParameterFilters.INPUTS):
            param, value = param_value.parameter, param_value.value

            if isinstance(param_value, InMemoryTarget):
                try:
                    param = param.update_value_meta_conf_from_runtime_value(
                        value, task_run.run.context.settings.tracking
                    )

                    task_run.tracker.log_parameter_data(
                        parameter=param,
                        target=param_value,
                        value=value,
                        operation_type=DbndTargetOperationType.read,
                        operation_status=DbndTargetOperationStatus.OK,
                    )
                except Exception as ex:
                    log_exception(
                        "Failed to log input param to tracking store.",
                        ex=ex,
                        non_critical=True,
                    )
    except Exception as ex:
        log_exception(
            "Failed to log input params to tracking store.", ex=ex, non_critical=True
        )


def _log_result(task_run, result):
    # type: (TaskRun, Any) -> None
    """
    For tracking mode. Logs the task result and adds it to the target_origin map to support relationships between
    dynamic tasks.
    """
    try:
        result_param = task_run.task.task_params.get_param_value(RESULT_PARAM)
        if not result_param:
            logger.debug(
                "No result params to log for task {}".format(task_run.task_af_id)
            )
            return

        # we now the parameter value is a target because this is an output param
        # the target is created in the task creation
        result_param_def, result_target = result_param.parameter, result_param.value

        # spread result into relevant fields.
        if isinstance(result_param_def, FuncResultParameter):
            # assign all returned values to relevant band Outputs
            if result is None:
                return

            for result_name, value in result_param_def.named_results(result):
                # we now the parameter value is a target because this is an output param
                # the target is created in the task creation
                parameter_value = task_run.task.task_params.get_param_value(result_name)

                _log_parameter_value(
                    task_run,
                    parameter_definition=parameter_value.parameter,
                    target=parameter_value.value,
                    value=value,
                )

        else:
            _log_parameter_value(
                task_run,
                parameter_definition=result_param_def,
                target=result_target,
                value=result,
            )

    except Exception as ex:
        log_exception(
            "Failed to log result to tracking store.", ex=ex, non_critical=True
        )


def _log_parameter_value(task_run, parameter_definition, target, value):
    # type: (TaskRun, ParameterDefinition, Target, Any) -> None
    # make sure it will be logged correctly
    parameter_definition = (
        parameter_definition.update_value_meta_conf_from_runtime_value(
            value, task_run.run.context.settings.tracking
        )
    )

    try:
        # case what if result is Proxy
        value_type = get_value_type_of_obj(value, parameter_definition.value_type)
        task_run.run.target_origin.add(target, value, value_type)
    except Exception as ex:
        log_exception(
            "Failed to register result to target tracking.", ex=ex, non_critical=True
        )

    try:
        task_run.tracker.log_parameter_data(
            parameter=parameter_definition,
            target=target,
            value=value,
            operation_type=DbndTargetOperationType.write,  # is it write? (or log?)
            operation_status=DbndTargetOperationStatus.OK,
        )
    except Exception as ex:
        log_exception(
            "Failed to log result to tracking store.", ex=ex, non_critical=True
        )



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._core.tracking.schemas.tracking_info_objects import TaskRunEnvInfo
from dbnd._vendor.marshmallow import fields, post_load


class TaskRunEnvInfoSchema(ApiStrictSchema):
    uid = fields.UUID()
    cmd_line = fields.String()

    user = fields.String()
    machine = fields.String()
    databand_version = fields.String()

    user_code_version = fields.String()
    user_code_committed = fields.Boolean()
    project_root = fields.String()

    user_data = fields.String()

    heartbeat = fields.DateTime()

    @post_load
    def make_object(self, data, **kwargs):
        return TaskRunEnvInfo(**data)



# © Copyright Databand.ai, an IBM Company 2022

"""
The abstract :py:class:`_BaseTask` class.
"""

import logging

from typing import TYPE_CHECKING, Optional

from dbnd._core.constants import TaskType
from dbnd._core.current import get_databand_context
from dbnd._core.errors import friendly_error
from dbnd._core.parameter.parameter_value import Parameters
from dbnd._core.utils.basics.nothing import NOTHING


if TYPE_CHECKING:
    from dbnd._core.settings import DatabandSettings
    from dbnd._core.task_build.task_decorator import TaskDecorator
    from dbnd._core.task_build.task_definition import TaskDefinition

logger = logging.getLogger(__name__)


class _BaseTask(object):
    _current_task_creation_id = 0

    # override to change get_task_family() -> changes task_family
    _conf__task_family = None
    _conf__track_source_code = True  # module/class/function user source code tracking

    _conf__task_type_name = TaskType.python
    _conf__task_ui_color = None
    #####
    # override output path format
    _conf__base_output_path_fmt = None

    _conf__tracked = True  # track task changes with TrackingStore

    # stores call spec for the @task definition
    task_decorator = None  # type: Optional[TaskDecorator]
    task_namespace = NOTHING

    @property
    def task_family(self):
        return self.task_definition.task_passport.task_family

    @property
    def _params(self):
        return self.task_params

    def __init__(
        self,
        task_name,
        task_definition,  # type: TaskDefinition
        task_signature_obj,
        task_params,  # type: Parameters
    ):
        super(_BaseTask, self).__init__()
        self.task_definition = task_definition
        self.task_params = task_params
        self.task_name = task_name  # type: str

        # miscellaneous
        self.task_id = "{}__{}".format(task_name, task_signature_obj.signature)
        self.task_type = self._conf__task_type_name

        self.task_signature_obj = task_signature_obj
        self.task_outputs_signature_obj = None

        # define it at the time of creation
        # we can remove the strong reference, the moment we have global cache for instances
        # otherwise if we moved to another databand_context, task_id relationships will not be found
        self.dbnd_context = get_databand_context()

        # we count task meta creation
        # even if cached task is going to be used we will increase creation id
        # if t2 created after t1, t2.task_creation_id > t1.task_creation_id
        _BaseTask._current_task_creation_id += 1
        self.task_creation_id = _BaseTask._current_task_creation_id

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.task_id == other.task_id

    def __hash__(self):
        return hash(self.task_id)

    def __repr__(self):
        return "%s" % self.task_id

    def __str__(self):
        return self.task_id

    @property
    def friendly_task_name(self):
        if self.task_name != self.task_family:
            return "%s[%s]" % (self.task_name, self.task_family)
        return self.task_name

    @property
    def settings(self):
        # type: () -> DatabandSettings

        return self.dbnd_context.settings

    @property
    def task_signature(self):
        # type: () -> str
        return self.task_signature_obj.signature

    @property
    def task_signature_source(self):
        # type: () -> str
        return self.task_signature_obj.signature_source

    def __iter__(self):
        raise friendly_error.task_build.iteration_over_task(self)

    def _task_banner(self, banner, verbosity):
        """
        customize task banner
        """
        return



# © Copyright Databand.ai, an IBM Company 2022

import logging

from functools import wraps
from time import time


logger = logging.getLogger(__name__)


def remote_debugger(port=15123):
    import pydevd

    pydevd.settrace("localhost", port=port, stdoutToServer=True, stderrToServer=True)


def log_call(name=None, log_f=None, print_start=False):
    log_f = log_f or logger.info

    def decorator(f):
        log_name = name or f.__name__

        @wraps(f)
        def wrapper(*args, **kwargs):
            if print_start:
                log_f("FUNC %s: starting", log_name)
            start = time()
            try:
                return f(*args, **kwargs)
            finally:
                end = time()
                log_f("FUNC %s: elapsed time: %s", log_name, end - start)

        return wrapper

    return decorator



# © Copyright Databand.ai, an IBM Company 2022

"""
Tracking store backends registry
"""

import logging

from typing import List, Optional

from dbnd._core.errors import friendly_error
from dbnd._core.tracking.backends import (
    CompositeTrackingStore,
    ConsoleStore,
    FileTrackingStore,
    TrackingStoreThroughChannel,
)
from dbnd._core.tracking.backends.tracking_store_composite import build_store_name


logger = logging.getLogger(__name__)

_BACKENDS_REGISTRY = {
    "file": FileTrackingStore,
    "console": ConsoleStore,
    "debug": TrackingStoreThroughChannel.build_with_console_debug_channel,
    ("api", "web"): TrackingStoreThroughChannel.build_with_web_channel,
    ("api", "async-web"): TrackingStoreThroughChannel.build_with_async_web_channel,
    ("api", "disabled"): TrackingStoreThroughChannel.build_with_disabled_channel,
}


def register_store(name, store_builder):
    if name in _BACKENDS_REGISTRY:
        raise Exception("Already registered")
    _BACKENDS_REGISTRY[name] = store_builder


def get_tracking_store(
    databand_ctx,
    tracking_store_names,
    api_channel_name,
    max_retires,
    tracker_raise_on_error,
    remove_failed_store,
):
    # type: (List[str], Optional[str], int, bool, bool) -> CompositeTrackingStore
    """
    Build a tracking stores based on the registry and wrap them with TrackingStoreComposite

    @param tracking_store_names: names of the wanted tracking stores
    @param api_channel_name: specify the name of the api channel if required
    @param max_retires: the amounts of retries to allowed the tracking store to run requests
    @param tracker_raise_on_error: True if the tracking store should raise fatal errors, False otherwise.
    @param remove_failed_store: True if a failed tracking store should be removed, False otherwise.
    @return:
    """

    tracking_store_instances = {}
    for name in tracking_store_names:
        if name == "api":
            name = (name, api_channel_name)

        tracking_store_builder = _BACKENDS_REGISTRY.get(name)
        if tracking_store_builder is None:
            raise friendly_error.config.wrong_store_name(name)

        instance = tracking_store_builder(databand_ctx)
        if isinstance(name, tuple):
            store_name = build_store_name(*name)
        else:
            store_name = name

        tracking_store_instances[store_name] = instance

    # We always use the CompositeTrackingStore cause it handle failure recovery for the inner tracking stores
    return CompositeTrackingStore(
        tracking_stores=tracking_store_instances,
        max_retires=max_retires,
        raise_on_error=tracker_raise_on_error,
        remove_failed_store=remove_failed_store,
    )



# patch.py -- For dealing with packed-style patches.
# Copyright (C) 2009-2013 Jelmer Vernooij <jelmer@jelmer.uk>
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Classes for dealing with git am-style patches.

These patches are basically unified diffs with some extra metadata tacked
on.
"""

from difflib import SequenceMatcher
import email.parser
import time

from dbnd._vendor.dulwich.objects import (
    Blob,
    Commit,
    S_ISGITLINK,
)

FIRST_FEW_BYTES = 8000


def write_commit_patch(f, commit, contents, progress, version=None, encoding=None):
    """Write a individual file patch.

    :param commit: Commit object
    :param progress: Tuple with current patch number and total.
    :return: tuple with filename and contents
    """
    encoding = encoding or getattr(f, "encoding", "ascii")
    if isinstance(contents, str):
        contents = contents.encode(encoding)
    (num, total) = progress
    f.write(
        b"From "
        + commit.id
        + b" "
        + time.ctime(commit.commit_time).encode(encoding)
        + b"\n"
    )
    f.write(b"From: " + commit.author + b"\n")
    f.write(
        b"Date: " + time.strftime("%a, %d %b %Y %H:%M:%S %Z").encode(encoding) + b"\n"
    )
    f.write(
        ("Subject: [PATCH %d/%d] " % (num, total)).encode(encoding)
        + commit.message
        + b"\n"
    )
    f.write(b"\n")
    f.write(b"---\n")
    try:
        import subprocess

        p = subprocess.Popen(
            ["diffstat"], stdout=subprocess.PIPE, stdin=subprocess.PIPE
        )
    except (ImportError, OSError):
        pass  # diffstat not available?
    else:
        (diffstat, _) = p.communicate(contents)
        f.write(diffstat)
        f.write(b"\n")
    f.write(contents)
    f.write(b"-- \n")
    if version is None:
        from dbnd._vendor.dulwich import __version__ as dulwich_version

        f.write(b"Dulwich %d.%d.%d\n" % dulwich_version)
    else:
        f.write(version.encode(encoding) + b"\n")


def get_summary(commit):
    """Determine the summary line for use in a filename.

    :param commit: Commit
    :return: Summary string
    """
    return commit.message.splitlines()[0].replace(" ", "-")


#  Unified Diff
def _format_range_unified(start, stop):
    'Convert range to the "ed" format'
    # Per the diff spec at http://www.unix.org/single_unix_specification/
    beginning = start + 1  # lines start numbering with one
    length = stop - start
    if length == 1:
        return "{}".format(beginning)
    if not length:
        beginning -= 1  # empty ranges begin at line just before the range
    return "{},{}".format(beginning, length)


def unified_diff(
    a, b, fromfile="", tofile="", fromfiledate="", tofiledate="", n=3, lineterm="\n"
):
    """difflib.unified_diff that can detect "No newline at end of file" as
    original "git diff" does.

    Based on the same function in Python2.7 difflib.py
    """
    started = False
    for group in SequenceMatcher(None, a, b).get_grouped_opcodes(n):
        if not started:
            started = True
            fromdate = "\t{}".format(fromfiledate) if fromfiledate else ""
            todate = "\t{}".format(tofiledate) if tofiledate else ""
            yield "--- {}{}{}".format(
                fromfile.decode("ascii"), fromdate, lineterm
            ).encode("ascii")
            yield "+++ {}{}{}".format(tofile.decode("ascii"), todate, lineterm).encode(
                "ascii"
            )

        first, last = group[0], group[-1]
        file1_range = _format_range_unified(first[1], last[2])
        file2_range = _format_range_unified(first[3], last[4])
        yield "@@ -{} +{} @@{}".format(file1_range, file2_range, lineterm).encode(
            "ascii"
        )

        for tag, i1, i2, j1, j2 in group:
            if tag == "equal":
                for line in a[i1:i2]:
                    yield b" " + line
                continue
            if tag in ("replace", "delete"):
                for line in a[i1:i2]:
                    if not line[-1:] == b"\n":
                        line += b"\n\\ No newline at end of file\n"
                    yield b"-" + line
            if tag in ("replace", "insert"):
                for line in b[j1:j2]:
                    if not line[-1:] == b"\n":
                        line += b"\n\\ No newline at end of file\n"
                    yield b"+" + line


def is_binary(content):
    """See if the first few bytes contain any null characters.

    :param content: Bytestring to check for binary content
    """
    return b"\0" in content[:FIRST_FEW_BYTES]


def shortid(hexsha):
    if hexsha is None:
        return b"0" * 7
    else:
        return hexsha[:7]


def patch_filename(p, root):
    if p is None:
        return b"/dev/null"
    else:
        return root + b"/" + p


def write_object_diff(f, store, old_file, new_file, diff_binary=False):
    """Write the diff for an object.

    :param f: File-like object to write to
    :param store: Store to retrieve objects from, if necessary
    :param old_file: (path, mode, hexsha) tuple
    :param new_file: (path, mode, hexsha) tuple
    :param diff_binary: Whether to diff files even if they
        are considered binary files by is_binary().

    :note: the tuple elements should be None for nonexistant files
    """
    (old_path, old_mode, old_id) = old_file
    (new_path, new_mode, new_id) = new_file
    old_path = patch_filename(old_path, b"a")
    new_path = patch_filename(new_path, b"b")

    def content(mode, hexsha):
        if hexsha is None:
            return Blob.from_string(b"")
        elif S_ISGITLINK(mode):
            return Blob.from_string(b"Submodule commit " + hexsha + b"\n")
        else:
            return store[hexsha]

    def lines(content):
        if not content:
            return []
        else:
            return content.splitlines()

    f.writelines(
        gen_diff_header((old_path, new_path), (old_mode, new_mode), (old_id, new_id))
    )
    old_content = content(old_mode, old_id)
    new_content = content(new_mode, new_id)
    if not diff_binary and (is_binary(old_content.data) or is_binary(new_content.data)):
        f.write(b"Binary files " + old_path + b" and " + new_path + b" differ\n")
    else:
        f.writelines(
            unified_diff(lines(old_content), lines(new_content), old_path, new_path)
        )


# TODO(jelmer): Support writing unicode, rather than bytes.
def gen_diff_header(paths, modes, shas):
    """Write a blob diff header.

    :param paths: Tuple with old and new path
    :param modes: Tuple with old and new modes
    :param shas: Tuple with old and new shas
    """
    (old_path, new_path) = paths
    (old_mode, new_mode) = modes
    (old_sha, new_sha) = shas
    yield b"diff --git " + old_path + b" " + new_path + b"\n"
    if old_mode != new_mode:
        if new_mode is not None:
            if old_mode is not None:
                yield ("old mode %o\n" % old_mode).encode("ascii")
            yield ("new mode %o\n" % new_mode).encode("ascii")
        else:
            yield ("deleted mode %o\n" % old_mode).encode("ascii")
    yield b"index " + shortid(old_sha) + b".." + shortid(new_sha)
    if new_mode is not None:
        yield (" %o" % new_mode).encode("ascii")
    yield b"\n"


# TODO(jelmer): Support writing unicode, rather than bytes.
def write_blob_diff(f, old_file, new_file):
    """Write blob diff.

    :param f: File-like object to write to
    :param old_file: (path, mode, hexsha) tuple (None if nonexisting)
    :param new_file: (path, mode, hexsha) tuple (None if nonexisting)

    :note: The use of write_object_diff is recommended over this function.
    """
    (old_path, old_mode, old_blob) = old_file
    (new_path, new_mode, new_blob) = new_file
    old_path = patch_filename(old_path, b"a")
    new_path = patch_filename(new_path, b"b")

    def lines(blob):
        if blob is not None:
            return blob.splitlines()
        else:
            return []

    f.writelines(
        gen_diff_header(
            (old_path, new_path),
            (old_mode, new_mode),
            (getattr(old_blob, "id", None), getattr(new_blob, "id", None)),
        )
    )
    old_contents = lines(old_blob)
    new_contents = lines(new_blob)
    f.writelines(unified_diff(old_contents, new_contents, old_path, new_path))


def write_tree_diff(f, store, old_tree, new_tree, diff_binary=False):
    """Write tree diff.

    :param f: File-like object to write to.
    :param old_tree: Old tree id
    :param new_tree: New tree id
    :param diff_binary: Whether to diff files even if they
        are considered binary files by is_binary().
    """
    changes = store.tree_changes(old_tree, new_tree)
    for (oldpath, newpath), (oldmode, newmode), (oldsha, newsha) in changes:
        write_object_diff(
            f,
            store,
            (oldpath, oldmode, oldsha),
            (newpath, newmode, newsha),
            diff_binary=diff_binary,
        )


def git_am_patch_split(f, encoding=None):
    """Parse a git-am-style patch and split it up into bits.

    :param f: File-like object to parse
    :param encoding: Encoding to use when creating Git objects
    :return: Tuple with commit object, diff contents and git version
    """
    encoding = encoding or getattr(f, "encoding", "ascii")
    encoding = encoding or "ascii"
    contents = f.read()
    if isinstance(contents, bytes) and getattr(email.parser, "BytesParser", None):
        parser = email.parser.BytesParser()
        msg = parser.parsebytes(contents)
    else:
        parser = email.parser.Parser()
        msg = parser.parsestr(contents)
    return parse_patch_message(msg, encoding)


def parse_patch_message(msg, encoding=None):
    """Extract a Commit object and patch from an e-mail message.

    :param msg: An email message (email.message.Message)
    :param encoding: Encoding to use to encode Git commits
    :return: Tuple with commit object, diff contents and git version
    """
    c = Commit()
    c.author = msg["from"].encode(encoding)
    c.committer = msg["from"].encode(encoding)
    try:
        patch_tag_start = msg["subject"].index("[PATCH")
    except ValueError:
        subject = msg["subject"]
    else:
        close = msg["subject"].index("] ", patch_tag_start)
        subject = msg["subject"][close + 2 :]
    c.message = (subject.replace("\n", "") + "\n").encode(encoding)
    first = True

    body = msg.get_payload(decode=True)
    lines = body.splitlines(True)
    line_iter = iter(lines)

    for line in line_iter:
        if line == b"---\n":
            break
        if first:
            if line.startswith(b"From: "):
                c.author = line[len(b"From: ") :].rstrip()
            else:
                c.message += b"\n" + line
            first = False
        else:
            c.message += line
    diff = b""
    for line in line_iter:
        if line == b"-- \n":
            break
        diff += line
    try:
        version = next(line_iter).rstrip(b"\n")
    except StopIteration:
        version = None
    return c, diff, version



# © Copyright Databand.ai, an IBM Company 2022

import attr


@attr.s
class TaskCallState(object):
    started = attr.ib(default=False)
    finished = attr.ib(default=False)
    result_saved = attr.ib(default=False)

    result = attr.ib(default=None)

    should_store_result = attr.ib(default=False)

    def start(self):
        self.started = True
        self.finished = False
        self.result = None

    def finish(self, result=None):
        self.finished = True
        if self.should_store_result:
            self.result_saved = True
            self.result = result



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.log.buffered_log_manager import BufferedLogManager


class BufferedMemoryHandler(logging.Handler):
    def __init__(self):
        logging.Handler.__init__(self)
        self.buffer_manager = BufferedLogManager()
        self.set_name("dbnd")

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            msg = self.format(record)
            self.buffer_manager.add_log_msg(msg)
        except RecursionError:  # See issue 36272
            raise
        except Exception:
            self.handleError(record)

    def get_log_body(self):
        return self.buffer_manager.get_log_body()



# © Copyright Databand.ai, an IBM Company 2022

import typing

from dbnd._core.task_build.task_signature import Signature


if typing.TYPE_CHECKING:
    from dbnd._core.task.task import Task


class TaskInstanceCache(object):
    def __init__(self):
        self.task_instances = {}
        self.task_obj_cache = {}

    # this is global values for now - we need to understand if it's really required
    # but if task has been created - it's something global, regardless current databand run
    def get_cached_task_obj(self, task_obj_cache_signature):
        # type: (Signature) -> Task
        return self.task_obj_cache.get(task_obj_cache_signature.signature, None)

    def get_task_by_id(self, task_id):
        # type: (str) -> Task
        return self.task_instances.get(task_id, None)

    def register_task_instance(self, task):
        # type: (Task) -> None
        self.task_instances[task.task_id] = task

    def register_task_obj_cache_instance(self, task, task_obj_cache_signature):
        self.task_obj_cache[task_obj_cache_signature.signature] = task



# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import logging
import typing

from contextlib import contextmanager

from dbnd._core.log.buffered_memory_handler import BufferedMemoryHandler
from dbnd._core.log.logging_utils import find_handler, redirect_stderr, redirect_stdout
from dbnd._core.settings import LocalEnvConfig
from dbnd._core.settings.log import LoggingConfig, _safe_is_typeof
from dbnd._core.task_run.log_preview import read_dbnd_log_preview
from dbnd._core.task_run.task_run_ctrl import TaskRunCtrl
from dbnd._core.tracking.dbnd_spark_init import attach_spark_logger, detach_spark_logger
from dbnd._core.utils.basics.environ_utils import environ_enabled
from targets.target_config import TargetConfig


if typing.TYPE_CHECKING:
    from targets import FileTarget

logger = logging.getLogger(__name__)

CURRENT_TASK_HANDLER_LOG = None


class TaskRunLogManager(TaskRunCtrl):
    def __init__(self, task_run):
        super(TaskRunLogManager, self).__init__(task_run)
        self.local_log_file = self.local_spark_log_file = None
        if self.run.is_orchestration:
            self.local_log_file = self.task_run.local_task_run_root.partition(
                name="%s.log" % task_run.attempt_number
            )  # type: FileTarget

            if environ_enabled("DBND__LOG_SPARK"):
                self.local_spark_log_file = self.task_run.local_task_run_root.partition(
                    name="%s-spark.log" % task_run.attempt_number
                )  # type: FileTarget
            else:
                self.local_spark_log_file = None

        self.remote_log_file = None
        if not isinstance(self.task.task_env, LocalEnvConfig):
            self.remote_log_file = self.task_run.attempt_folder.partition(
                name=str(task_run.attempt_number),
                config=TargetConfig().as_file().txt,
                extension=".log",
            )  # type: FileTarget

        # file handler for task log
        # if set -> we are in the context of capturing
        self._log_task_run_into_file_active = False

    @contextmanager
    def capture_stderr_stdout(self, logging_target=None):
        #  redirecting all messages from sys.stderr/sys.stdout into logging_target
        # WARNING: this can create stdout/stderr loop if :
        #     we redirect into logger, and logger is writing to current sys.stdout that is current redirect
        #     ( not the original one before the wrapping)
        #    -> LOOP:   some print ->  redirect_stdout() -> logger -> print -> redirect_stdout() ...
        if not self.task.settings.log.capture_stdout_stderr:
            yield None
            return

        airflow_root_console_handler = find_handler(logging.root, "console")
        if _safe_is_typeof(airflow_root_console_handler, "RedirectStdHandler"):
            yield None
            return

        with contextlib.ExitStack() as stack:
            logging_target_stdout = logging_target or logging.getLogger("dbnd.stdout")
            if logging_target_stdout.isEnabledFor(logging.INFO):
                stack.enter_context(
                    redirect_stdout(logging_target_stdout, logging.INFO)
                )

            logging_target_stderr = logging_target or logging.getLogger("dbnd.stderr")
            if logging_target_stderr.isEnabledFor(logging.WARN):
                stack.enter_context(
                    redirect_stderr(logging_target_stderr, logging.WARN)
                )

            yield

    @contextmanager
    def capture_task_log(self):
        global CURRENT_TASK_HANDLER_LOG
        log_file = self.local_log_file

        log_settings: LoggingConfig = self.task.settings.log
        if self._log_task_run_into_file_active or not log_settings.capture_task_run_log:
            yield None
            return

        if (
            # there is no really any good way to find if this is runtime tracking for airflow operator
            # should be refactored as soon as possible
            self.task.task_family.endswith("_execute")
        ):
            yield None
            return

        if not log_file or not self.run.is_orchestration:
            handler = log_settings.get_task_log_memory_handler()
        else:
            handler = log_settings.get_task_log_file_handler(log_file)
            logger.debug("Capturing task log into '%s'", log_file)

        if not handler:
            yield None
            return
        target_logger = logging.root

        try:
            attach_spark_logger(self.local_spark_log_file)
            target_logger.addHandler(handler)
            self._log_task_run_into_file_active = True
            CURRENT_TASK_HANDLER_LOG = handler

            with self.capture_stderr_stdout():
                yield handler
        except Exception as task_ex:
            CURRENT_TASK_HANDLER_LOG = None
            raise task_ex
        finally:
            detach_spark_logger(self.local_spark_log_file)
            try:
                target_logger.removeHandler(handler)
                handler.close()
            except Exception:
                logger.warning("Failed to close file handler for log %s", log_file)

            self._log_task_run_into_file_active = False
            self._upload_task_log_preview(handler)

    def _upload_task_log_preview(self, log_handler):
        try:
            if isinstance(log_handler, logging.FileHandler):
                log_body = self.read_log_body()
            elif isinstance(log_handler, BufferedMemoryHandler):
                log_body = log_handler.get_log_body()
            else:
                logger.exception("unable to read log body for %s", self)
                return

        except Exception as read_log_err:
            logger.exception("failed to read log preview for %s:%s", self, read_log_err)
        else:
            try:
                self.save_log_preview(log_body)
            except Exception as save_log_err:
                logger.exception(
                    "failed to save task run log preview for %s:%s", self, save_log_err
                )

        self.write_remote_log()

    def read_log_body(self):
        try:
            spark_log_file = (
                self.local_spark_log_file.path
                if environ_enabled("DBND__LOG_SPARK")
                else None
            )
            return read_dbnd_log_preview(self.local_log_file.path, spark_log_file)

        except Exception as ex:
            logger.warning(
                "Failed to read log (%s) for %s: %s",
                self.local_log_file.path,
                self.task,
                ex,
            )
            return None

    def write_remote_log(self):
        if (
            self.task.settings.log.remote_logging_disabled
            or self.remote_log_file is None
            or self.local_log_file is None
        ):
            return

        try:
            self.remote_log_file.copy_from_local(self.local_log_file.path)

        except Exception as ex:
            logger.warning(
                "Failed to upload the log from {local_path} "
                "to the remote log at {remote_log} "
                "for task {task}: {exception}".format(
                    local_path=self.local_log_file.path,
                    remote_log=self.remote_log_file.path,
                    task=self.task,
                    exception=ex,
                )
            )

    def save_log_preview(self, log_body):
        self.task_run.tracker.save_task_run_log(log_body, self.local_log_file)



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import logging
import os
import subprocess

import six

from dbnd._core.errors.friendly_error.task_execution import failed_to_run_cmd
from dbnd._core.log.logging_utils import raw_log_formatting


logger = logging.getLogger(__name__)


def _print_log(msg):
    print(msg, end="")


def run_cmd(
    cmd,
    name="process",
    env=None,
    stdout_handler=_print_log,
    return_code=0,
    shell=False,
    **kwargs
):
    if env:
        os_env = os.environ.copy()
        os_env.update(env)
        kwargs["env"] = os_env

    sp = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=-1,
        universal_newlines=True,
        shell=shell,  # nosec B602 (Ignore this for bandit as this is code used for orchestration)
        **kwargs
    )

    with raw_log_formatting():
        for l in iter(sp.stdout.readline, ""):
            stdout_handler(l)
    returned_code = sp.wait()

    if isinstance(cmd, six.string_types):
        cmd_str = cmd
    else:
        cmd_str = subprocess.list2cmdline(cmd)

    # Check spark-submit return code. In Kubernetes mode, also check the value
    # of exit code in the log, as it may differ.
    if return_code is not None and returned_code != return_code:
        raise failed_to_run_cmd(name=name, cmd_str=cmd_str, return_code=returned_code)



# © Copyright Databand.ai, an IBM Company 2022


def generate_airflow_cmd(dag_id, task_id, execution_date, is_root_task=False):
    return "airflow {sub_command} {dag_id}{task_id} {start_date} {end_date}".format(
        sub_command="backfill" if is_root_task else "test",
        dag_id=dag_id,
        task_id=" %s" % task_id if not is_root_task else "",
        start_date="-s %s" % execution_date,
        end_date="-e %s" % execution_date,
    )


def generate_airflow_func_call(
    dag_id,
    schedule_interval,
    execution_date,
    task_type=None,
    task_id=None,
    task_retries=None,
    task_command=None,
    is_root_task=False,
):
    if is_root_task:
        return (
            "DAG(dag_id='{dag_id}', default_args=args, schedule_interval='{schedule_interval}')"
            ".run(start_date={start_date}, end_date={end_date}, execution_date={execution_date})".format(
                dag_id=dag_id,
                schedule_interval=schedule_interval,
                start_date=execution_date,
                end_date=execution_date,
                execution_date=execution_date,
            )
        )
    elif task_id is not None:
        return (
            "{operator}(task_id='{task_id}', retries='{retries}', {command}))".format(
                operator=task_type,
                task_id=task_id,
                retries=task_retries,
                command=task_command,
            )
        )



# © Copyright Databand.ai, an IBM Company 2022

import os

from dbnd import dbnd_config, relative_path


TEST_CONF_FILE_NAME = "databand-test.cfg"


def _add_local_test_config(calling_file):
    local_test_file = relative_path(calling_file, TEST_CONF_FILE_NAME)
    if os.path.exists(local_test_file):
        dbnd_config.set_from_config_file(local_test_file)


def _add_global_test_config():
    curent_dir = os.path.normpath(os.getcwd())
    current_dir_split = curent_dir.split(os.sep)
    current_dir_split_reversed = reversed(list(enumerate(current_dir_split)))

    # traversing all the way up to until finding relevant anchor files
    for index, current_folder in current_dir_split_reversed:
        current_path = os.path.join("/", *current_dir_split[1 : (index + 1)])
        global_config_file = os.path.join(current_path, TEST_CONF_FILE_NAME)
        if os.path.exists(global_config_file):
            dbnd_config.set_from_config_file(global_config_file)
            break


def add_test_configuration(calling_file):
    _add_local_test_config(calling_file)
    _add_global_test_config()



# © Copyright Databand.ai, an IBM Company 2022

from targets import InMemoryTarget


class InlineTarget(InMemoryTarget):
    """
    represents in-place value provided by user,
    used by inline functions calls to track data dependencies
    """

    def __init__(self, root_target, obj, value_type, **kwargs):
        super(InlineTarget, self).__init__(obj=obj, value_type=value_type, **kwargs)
        self.root_target = root_target
        self.target_meta = root_target.target_meta

    def __repr__(self):
        if self.root_target:
            return repr(self.root_target)
        return super(InlineTarget, self).__repr__()

    def __str__(self):
        if self.root_target:
            return str(self.root_target)
        return super(InlineTarget, self).__str__()



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Any

from pandas import DataFrame

from targets.extras import DataTargetCtrl
from targets.marshalling import get_marshaller_ctrl
from targets.target_config import FileFormat, file
from targets.utils.performance import target_timeit


logger = logging.getLogger(__name__)
file_table = file.with_format(FileFormat.table)


class PandasMarshallingCtrl(DataTargetCtrl):
    def __init__(self, target):
        super(PandasMarshallingCtrl, self).__init__(target)

    def read(self, config=None, **kwargs):
        # type: (file, **Any) -> 'DataFrame'
        pd_m = get_marshaller_ctrl(self.target, DataFrame, config=config)
        return pd_m.load(**kwargs)

    def read_partitioned(self, config=None, **kwargs):
        pd_m = get_marshaller_ctrl(self.target, DataFrame, config=config)
        for t in pd_m.load_partitioned(**kwargs):
            yield t

    @target_timeit
    def to(self, df, config=None, **kwargs):
        pd_m = get_marshaller_ctrl(self.target, value_type=DataFrame, config=config)
        return pd_m.dump(df, **kwargs)

    def read_csv(self, **kwargs):
        return self.read(config=file.csv, **kwargs)

    def read_table(self, **kwargs):
        return self.read(config=file_table, **kwargs)

    def read_hdf(self, **kwargs):
        return self.read(config=file.hdf5, **kwargs)

    def read_json(self, **kwargs):
        return self.read(config=file.json, **kwargs)

    def read_feather(self, **kwargs):
        return self.read(config=file.feather, **kwargs)

    def read_parquet(self, **kwargs):
        return self.read(config=file.parquet, **kwargs)

    def read_pickle(self, **kwargs):
        return self.read(config=file.pickle, **kwargs)

    def to_csv(self, df, **kwargs):
        return self.to(df, config=file.csv, **kwargs)

    def to_json(self, df, **kwargs):
        return self.to(df, config=file.json, **kwargs)

    def to_hdf(self, df, **kwargs):
        return self.to(df, config=file.hdf5, **kwargs)

    def to_feather(self, df, **kwargs):
        return self.to(df, config=file.feather, **kwargs)

    def to_table(self, df, **kwargs):
        return self.to(df, config=file_table, **kwargs)

    def to_parquet(self, df, **kwargs):
        return self.to(df, config=file.parquet, **kwargs)

    def to_pickle(self, df, **kwargs):
        return self.to(df, config=file.pickle, **kwargs)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from typing import List

import six

from dbnd._core.constants import RESULT_PARAM
from dbnd._core.errors import DatabandBuildError, friendly_error
from dbnd._core.parameter import get_parameter_for_value_type
from dbnd._core.parameter.parameter_builder import (
    PARAMETER_FACTORY,
    ParameterFactory,
    build_parameter,
    parameter,
)
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    _ParameterKind,
)
from dbnd._core.task_build.task_results import FuncResultParameter
from dbnd._core.utils.basics.nothing import NOTHING, is_defined, is_not_defined
from dbnd._core.utils.callable_spec import (
    guess_func_arg_value_type,
    guess_func_return_type,
)
from targets.values import get_value_type_of_type
from targets.values.builtins_values import DefaultObjectValueType


if typing.TYPE_CHECKING:
    from dbnd._core.task_build.task_decorator import TaskDecorator
    from dbnd._core.utils.callable_spec import CallableSpec
logger = logging.getLogger(__name__)


class DecoratedCallableParamBuilder(object):
    def __init__(self, base_params, task_decorator):
        self.base_params = base_params

        self.task_decorator = task_decorator  # type: TaskDecorator

        self.callable_spec = task_decorator.get_callable_spec()  # type: CallableSpec
        self._context = task_decorator.task_passport.full_task_family
        self.exclude = {RESULT_PARAM, "self"}

        self.decorator_kwargs_params = {}
        self.func_spec_params = {}
        self.result_params = {}

    def _build_func_spec_params(self, decorator_kwargs_params):
        params = {}
        # let go over all kwargs of the functions
        for k in self.callable_spec.args:
            if (
                k in self.exclude or k in decorator_kwargs_params
            ):  # excluded or processed already
                continue
            context = self._get_param_context(k)

            default = self.callable_spec.defaults.get(k, NOTHING)
            if isinstance(default, ParameterFactory):
                # it's inplace defition
                # user_param= parameter[str]
                params[k] = build_parameter(default, context=context)
            elif k in self.base_params:
                # we have param definition from "base class"
                # so we just need to provide a default value for the parameter
                # otherwise - do nothing, as we are good
                if is_defined(default):
                    params[k] = default

            else:
                try:
                    # regular value
                    param_value_type = guess_func_arg_value_type(
                        self.callable_spec, k, default
                    )
                    if param_value_type is None:
                        # fallback to "object"
                        param_value_type = DefaultObjectValueType()

                    param = get_parameter_for_value_type(param_value_type)
                    param = param.default(default)
                    params[k] = build_parameter(param, context=context)
                except Exception:
                    logger.exception("Failed to analyze function arg %s", context)
                    raise

        if self.callable_spec.varargs:
            # create a param with the name of `*args` argument of the function
            params[self.callable_spec.varargs] = build_parameter(
                parameter[list], context=self._get_param_context("*args")
            )

        if self.callable_spec.varkw:
            # create a param with the name of `**kwargs` argument of the function
            params[self.callable_spec.varkw] = build_parameter(
                parameter[typing.Dict[str, typing.Any]],
                context=self._get_param_context("**kwargs"),
            )
        return params

    def _build_decorator_kwargs_params(self):
        params = {}
        for k, param in six.iteritems(self.task_decorator.decorator_kwargs):
            if k in self.exclude:  # we'll take care of result param later
                continue

            if param is None:
                self.exclude.add(k)
                continue

            context = self._get_param_context(k)

            if k not in self.callable_spec.args and k not in self.base_params:
                # we have parameter which is not part of real function signature
                # @task(some_unknown_parameter=parameter)
                logger.info(
                    "{} is not part of parameters, creating hidden parameter".format(
                        context
                    )
                )

            if k in self.callable_spec.defaults:
                if isinstance(self.callable_spec.defaults[k], ParameterFactory):
                    raise DatabandBuildError(
                        "{}: {} has conlficted definition in function and in decorator itself".format(
                            context, k
                        )
                    )
                if is_defined(param.parameter.default):
                    logger.warning(
                        "Default value conflict between function and @task decorator"
                    )
                param = param.default(self.callable_spec.defaults.get(k))

            if k not in self.base_params or isinstance(param, ParameterFactory):
                # we are going to build a new parameter
                param = build_parameter(param, context=context)
            params[k] = param

        return params

    def _build_multiple_outputs_result(self, result_deco_spec):
        # type: (List[ParameterFactory]) -> (List[ParameterDefinition], ParameterDefinition)

        context = self._get_param_context(RESULT_PARAM)
        res = []
        for i, lv in enumerate(result_deco_spec):
            lv = build_parameter(lv, context="%s.%s" % (context, i))

            if not lv.name:
                raise DatabandBuildError(
                    "{}[{}]: {} should have name".format(context, i, type(lv))
                )
            if not lv.is_output():
                raise DatabandBuildError(
                    "{}[{}]: {} should marked as output".format(context, i, type(lv))
                )

            res.append(lv)

        param = FuncResultParameter(
            schema=res,
            value_type=DefaultObjectValueType(),
            name=RESULT_PARAM,
            significant=False,
            kind=_ParameterKind.task_output,
        )
        return param

    def _get_result_parameter_part(self, p, name_hint, value_type_hint):
        if isinstance(p, six.string_types):
            # we got a string  - it's name of parameter
            name_hint = p
            p = None
        elif not isinstance(p, ParameterFactory):
            # we got a type  - it's inline type of parameter
            value_type_hint = get_value_type_of_type(p, inline_value_type=True)
            p = None
        else:
            pass

        if p is None:
            if value_type_hint:
                # we create based on value type
                p = get_parameter_for_value_type(value_type_hint)
            else:
                p = PARAMETER_FACTORY

        if p.parameter.name is None and name_hint:
            p = p.modify(name=name_hint)

        if p.parameter.value_type is None and value_type_hint:
            p = p.modify(value_type=value_type_hint)

        return p.output

    def _get_result_parameter(self):
        context = self._get_param_context(RESULT_PARAM)

        return_spec = guess_func_return_type(self.callable_spec)

        deco_spec = None
        # first of all , let parse the definition we have
        if RESULT_PARAM in self.task_decorator.decorator_kwargs:
            # @task(result=...)
            deco_spec = self.task_decorator.decorator_kwargs[RESULT_PARAM]
            if isinstance(deco_spec, dict):
                raise friendly_error.task_parameters.dict_in_result_definition(
                    deco_spec
                )

            # @task(result=None)
            if deco_spec is None:
                # user explicitly don't want to have result value
                return {}

            if isinstance(deco_spec, six.string_types):
                # we have result = "output1,output2"
                # support both space and comma
                deco_spec = deco_spec.replace(",", " ").split()
                if len(deco_spec) == 1:
                    deco_spec = deco_spec[0]
            elif isinstance(deco_spec, tuple):
                deco_spec = list(deco_spec)

            # user didn't specify - so we don't have any "hints"
            if is_not_defined(return_spec):
                return_spec = None
            elif return_spec is not None:
                # we will use type hints from  "-> ..." spec only if it's has exact match to our params
                return_spec = self._validate_return_spec(deco_spec, return_spec)
        else:
            # we don't have @task(result=)
            if return_spec is None:
                # .. -> None
                # user explicitly don't want to have result value
                return {}
            # let return default parameter ( pickle in @task)
            if is_not_defined(return_spec):
                return build_parameter(self.task_decorator.task_default_result, context)

            # so we have something in return speck, let use it
            if isinstance(return_spec, list):
                # we can get names from ->
                deco_spec = [r[0] for r in return_spec]
            else:
                # or we just use default  name
                deco_spec = RESULT_PARAM

        # so now we have 2 cases
        # 1. we have list of results -->
        if isinstance(deco_spec, list):
            result = []
            for i, deco_p in enumerate(deco_spec):
                value_type_hint = None
                if return_spec:
                    _, value_type_hint = return_spec[i]

                deco_p = self._get_result_parameter_part(
                    p=deco_p, name_hint="result_%s" % i, value_type_hint=value_type_hint
                )
                result.append(deco_p)
            param = self._build_multiple_outputs_result(result)

        # 2. we have only one result-->
        else:
            param = self._get_result_parameter_part(
                p=deco_spec, name_hint=RESULT_PARAM, value_type_hint=return_spec
            )
        return build_parameter(param, context)

    def build_func_params(self):
        if not self.callable_spec:
            return

        # calc auto parameters from function spec
        # it's easy, as we take function value or deco value.
        self.decorator_kwargs_params = self._build_decorator_kwargs_params()
        self.func_spec_params = self._build_func_spec_params(
            self.decorator_kwargs_params
        )
        self.result_params = self._build_func_result_params()

        self._validate_result_params()

    def _validate_result_params(self):
        conflict_keys = set(self.result_params.keys()).intersection(
            set(self.decorator_kwargs_params.keys()) | set(self.func_spec_params.keys())
        )
        if conflict_keys:
            raise friendly_error.task_parameters.result_and_params_have_same_keys(
                self._context, conflict_keys
            )

    def _build_func_result_params(self):
        # much more complicated , we need to "smart merge" the result
        # and we need the result type
        result_param = self._get_result_parameter()
        result_params = {}
        if result_param:
            result_params[RESULT_PARAM] = result_param
            if isinstance(result_param, FuncResultParameter):
                result_params.update({p.name: p for p in result_param.schema})
        return result_params

    def _validate_return_spec(self, deco_spec, return_spec):
        # validate lengths

        def err(msg):
            logger.warning(
                "%s() return type definitions are different: \n\t"
                "@task(result='%s')\n\t"
                "# type: -> '%s'\n\t error:%s\n\t"
                " @task definition will be used!",
                self._context,
                deco_spec,
                return_spec,
                msg,
            )

        deco_is_list = isinstance(deco_spec, list)
        return_spec_is_list = isinstance(return_spec, list)
        if not deco_is_list:
            if not return_spec_is_list:
                # just one definition, good!
                return return_spec

            err("function return definition is a list")
            return None

        # deco is list
        if not return_spec_is_list:
            err("function return definition is a single type")
            return None

        if len(deco_spec) != len(return_spec):
            err("number of outputs is different")
            return None
        return return_spec

    def _get_param_context(self, param_name):
        return "%s.%s" % (self._context, param_name)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import time

from functools import wraps

from dbnd._core.current import get_target_logging_level
from dbnd._core.utils.seven import contextlib


logger = logging.getLogger(__name__)


def target_timeit(method):
    @wraps(method)
    def wrapper(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        state_obj = args[0]
        if hasattr(state_obj, "target"):
            state_obj = state_obj.target
        if hasattr(state_obj, "path"):
            state_obj = state_obj.path

        logger.debug("%r %s %2.2f ms" % (method.__name__, state_obj, (te - ts) * 1000))
        return result

    return wrapper


@contextlib.contextmanager
def target_timeit_log(target, operation):
    from dbnd import log_metric

    log_level = get_target_logging_level()

    path = target
    if hasattr(target, "target"):
        path = target.target
    if hasattr(target, "path"):
        path = target.path

    name = target.source.name if target.source else path

    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        delta_ms = (end_time - start_time) * 1000
        logger.log(
            level=log_level,
            msg="Total {} time for target {} is {} milliseconds".format(
                operation, path, delta_ms
            ),
        )
        log_metric("marshalling_{}".format(name), delta_ms, source="system")



# © Copyright Databand.ai, an IBM Company 2022

import functools
import logging
import os
import re

from collections.abc import Mapping
from configparser import ConfigParser
from typing import Any

import attr
import six

from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.configuration.config_store import _ConfigStore
from dbnd._core.configuration.config_value import ConfigValue, ConfigValuePriority
from dbnd._core.configuration.environ_config import (
    get_dbnd_custom_config,
    get_dbnd_environ_config_file,
    in_quiet_mode,
    is_unit_test_mode,
)
from dbnd._core.errors import DatabandConfigError
from dbnd._core.utils.project.project_fs import (
    databand_config_path,
    databand_system_path,
)
from dbnd._vendor.snippets.airflow_configuration import expand_env_var
from targets import LocalFileSystem, target
from targets.pipes import Text


logger = logging.getLogger(__name__)
_DBND_ENVIRON_RE = re.compile(r"^DBND__([\w\d_\-]+)__([\w\d_\-]+)$")
_SECTION_NAME_RE = re.compile(r"(.*)\.([^.]*)")


def _default_configuration_paths():
    # we always have "library config"
    yield databand_config_path("databand-core.cfg")

    system_config = os.path.expanduser("/etc/databand.cfg")
    if os.path.isfile(system_config):
        yield system_config

    # now we can start to look for project configs
    dbnd_project_config = get_dbnd_project_config()

    possible_locations = [
        databand_system_path("databand-system.cfg"),
        dbnd_project_config.dbnd_project_path("conf", "databand.cfg"),  # deprecated
        dbnd_project_config.dbnd_project_path("databand.cfg"),  # deprecated
        get_dbnd_custom_config(),
        dbnd_project_config.dbnd_project_path("project.cfg"),
    ]
    env_config = get_dbnd_environ_config_file()
    if env_config:
        possible_locations.append(env_config)
    for value in possible_locations:
        value = expand_env_var(value)
        if os.path.isfile(value):
            yield value

    user_config = expand_env_var("~/.dbnd/databand.cfg")
    if os.path.isfile(user_config):
        yield user_config

    if is_unit_test_mode():
        tests_config_path = databand_system_path("databand-test.cfg")
        if os.path.exists(tests_config_path):
            yield tests_config_path


def read_from_config_stream(config_fp, source="<stream>"):
    """
    Read config from config file (.ini, .cfg)
    """
    parser = ConfigParser()
    parser._read(config_fp, source)

    source = "config[{file_name}]".format(file_name=os.path.basename(str(source)))
    new_config = _ConfigStore()
    for section in parser.sections():
        for option in parser.options(section):
            value = parser.get(section, option)
            new_config.set_config_value(
                section, option, ConfigValue(value, source, require_parse=True)
            )

    return new_config


def read_from_config_file(config_file_target):
    """
    Read config from config file (.ini, .cfg)
    """

    # Check if config files are inside zip first - can happen if we run in fat wheel
    if (
        isinstance(config_file_target.fs, LocalFileSystem)
        and ".zip/" in config_file_target.path
    ):
        zip_file_path, config_path_inside_zip = config_file_target.path.split(".zip/")
        if os.path.exists(zip_file_path + ".zip"):
            import zipfile

            archive = zipfile.ZipFile(zip_file_path + ".zip", "r")
            if config_path_inside_zip not in archive.namelist():
                raise DatabandConfigError(
                    "Failed to read configuration file at %s, file not found!"
                    % config_file_target
                )
            archive = zipfile.ZipFile(zip_file_path + ".zip", "r")
            with archive.open(config_path_inside_zip) as file_io:
                return read_from_config_stream(
                    Text.pipe_reader(file_io), str(config_file_target)
                )

    if not config_file_target.exists():
        raise DatabandConfigError(
            "Failed to read configuration file at %s, file not found!"
            % config_file_target
        )
    with config_file_target.open("r") as fp:
        return read_from_config_stream(fp, str(config_file_target))


def read_from_config_files(config_files):
    files_to_load = [target(f) for f in config_files]
    configs = []

    if not in_quiet_mode():
        logger.info(
            "Reading configuration from: \n\t%s\n", "\n\t".join(map(str, files_to_load))
        )

    for f in files_to_load:
        try:
            configs.append(read_from_config_file(f))
        except DatabandConfigError:
            raise
        except Exception as ex:
            raise DatabandConfigError(
                "Failed to read configuration file at %s: %s" % (f, ex),
                nested_exceptions=ex,
            )

    merged_file_config = functools.reduce((lambda x, y: x.update(y)), configs)
    return merged_file_config


def read_environ_config():
    """
    Read configuration from process environment
    Every env var in following format will be added to config
    $DBND__SECTION__KEY=value  (please notice double underscore "__")
    :return:
    """
    return get_environ_config_from_dict(os.environ, "environ")


def get_environ_config_from_dict(env_dict, source_prefix):
    dbnd_environ = _ConfigStore()
    for key, value in six.iteritems(env_dict):
        if not key.startswith("DBND__"):
            continue
        # must have format DBND__{SECTION}__{KEY} (note double underscore)
        dbnd_key_var = _DBND_ENVIRON_RE.match(key)
        if dbnd_key_var:
            section, key = dbnd_key_var.group(1), dbnd_key_var.group(2)
            dbnd_environ.set_config_value(
                section,
                key,
                ConfigValue(
                    value,
                    source="{}[{}]".format(source_prefix, key),
                    require_parse=True,
                ),
            )
        else:
            # check that it's known name, or print error
            pass

    return dbnd_environ


def parse_and_build_config_store(
    source,
    config_values,
    auto_section_parse=False,
    priority=None,
    override=False,  # might be deprecated in favor of priority
    extend=False,
):
    # type:(str, Mapping[str, Mapping[str, Any]], bool, bool , bool, bool)->_ConfigStore
    """
    Read user defined values. Following format are supported:
        1. SomeTask.some_param [ParameterDefinition] : value
        2. { "section" : { "key" : "value" }}
        3 ? "SomeTask.some_param" [str]  : value
    """
    if isinstance(config_values, _ConfigStore):
        return config_values

    from dbnd._core.parameter.parameter_definition import ParameterDefinition

    new_config = _ConfigStore()
    new_config.source = source
    if not config_values:
        return new_config

    for section, section_values in six.iteritems(config_values):
        if isinstance(section, six.string_types):
            if auto_section_parse:
                m = _SECTION_NAME_RE.match(section)
                if m:  # section contains key!
                    section, key = m.group(1), m.group(2)
                    section_values = {key: section_values}

            if not isinstance(section_values, Mapping):
                raise DatabandConfigError(
                    "can't convert '%s' to configuration " % config_values
                )
        elif isinstance(section, ParameterDefinition):
            # this is parameter ->  Spark.jars = ["jars"]
            section_values = {section.name: section_values}
            section = section.task_config_section

        else:
            raise Exception("section='%s' not supported" % section)

        new_section = new_config[section]
        for key, value in six.iteritems(section_values):
            if key in new_section:
                raise Exception(
                    "multiple definition of {section}.{key} at {config}".format(
                        section=section, key=key, config=config_values
                    )
                )
            if isinstance(key, ParameterDefinition):
                key = key.name
            if not isinstance(value, ConfigValue):
                if priority is None:
                    priority = (
                        ConfigValuePriority.OVERRIDE
                        if override
                        else ConfigValuePriority.NORMAL
                    )
                value = ConfigValue(
                    value=value,
                    source=source,
                    require_parse=False,
                    priority=priority,
                    extend=extend,
                )
            else:
                # we can have override values without source
                if value.source is None:
                    value = attr.evolve(value, source=source)

                if extend:
                    value = attr.evolve(value, extend=extend)

            new_config.set_config_value(section, key, value)

    return new_config



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.utils.seven import contextlib


# noinspection PyBroadException
@contextlib.contextmanager
def nested(*managers):
    """Combine multiple context managers into a single nested context manager.

    The one advantage of this function over the multiple manager form of the
    with statement is that argument unpacking allows it to be
    used with a variable number of context managers as follows:

       with nested(*managers):
           do_something()

    """
    if not managers:
        yield
        return

    with contextlib.ExitStack() as stack:
        for mgr in managers:
            if mgr is None:
                continue
            stack.enter_context(mgr)
        yield



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.constants import AD_HOC_DAG_PREFIX


def dont_track(obj):
    """
    Will not track obj.

    support airflow operators, DAGs, and functions.
    allows excluding from tracking when using dbnd-airflow-auto-tracking, track_modules, and track_dag.
    can be used as a function and decorator as shown in examples below.
    As a function::

        dag = DAG()
        with dag:
            operator = PythonOperator()

        dont_track(operator)
        dont_track(dag)

    As a decorator::

        @dont_track
        def f():
            pass
    """
    obj._dont_track = True
    return obj


def should_not_track(obj):
    if hasattr(obj, "dag"):
        if getattr(obj.dag, "_dont_track", False):
            return True
        dag_id = getattr(obj.dag, "dag_id", None)
        if dag_id and dag_id.startswith(AD_HOC_DAG_PREFIX):
            return True
    return getattr(obj, "_dont_track", False)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import queue
import threading

from time import sleep, time
from typing import Any, Dict

import attr

from dbnd._core.current import in_tracking_run, is_orchestration_run
from dbnd._core.errors.base import DatabandWebserverNotReachableError
from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.log.external_exception_logging import log_exception_to_server
from dbnd._core.tracking.backends.abstract_tracking_store import is_state_call
from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel
from dbnd._core.tracking.backends.channels.tracking_web_channel import (
    TrackingWebChannel,
)
from dbnd._core.tracking.backends.tracking_store_composite import try_run_handler
from dbnd._vendor.pendulum import utcnow


logger = logging.getLogger(__name__)

_TERMINATOR = object()


class TrackingAsyncWebChannelBackgroundWorker(object):
    def __init__(self, item_processing_handler, skip_processing_callback):
        self.item_processing_handler = item_processing_handler
        self.skip_processing_callback = skip_processing_callback

        self._lock = None
        self._queue = None
        self._thread = None
        self._thread_for_pid = None

    @property
    def queue(self):
        if self._queue is None:
            self._queue = queue.Queue()
        return self._queue

    @property
    def lock(self):
        if self._lock is None:
            self._lock = threading.Lock()
        return self._lock

    @property
    def is_alive(self) -> bool:
        if self._thread_for_pid != os.getpid():
            return False
        if not self._thread:
            return False
        return self._thread.is_alive()

    def start(self) -> None:
        with self.lock:
            if not self.is_alive:
                self._thread = threading.Thread(
                    target=self._thread_worker,
                    daemon=True,
                    name="dbnd.TrackingAsyncWebChannelBackgroundWorker",
                )
                self._thread.start()
                self._thread_for_pid = os.getpid()

    def flush(self, timeout) -> None:
        logger.debug("background worker got flush request")
        with self.lock:
            if self.is_alive:
                self.queue.put(_TERMINATOR)
                self._wait_flush(timeout)
                self._thread = None
                self._thread_for_pid = None
                self._lock = None
                self._queue = None

    def _wait_flush(self, timeout: float) -> None:
        initial_timeout = min(0.1, timeout)
        if not self._timed_queue_join(initial_timeout):
            logger.debug("%d event(s) pending on flush", self.get_qsize())

            if not self._timed_queue_join(timeout - initial_timeout):
                raise TimeoutError(
                    "Flush timed out and dropped some events", self.get_qsize(), timeout
                )

    def _timed_queue_join(self, timeout: float) -> bool:
        deadline = time() + timeout

        self.queue.all_tasks_done.acquire()

        try:
            while self.queue.unfinished_tasks:
                delay = deadline - time()
                if delay <= 0:
                    return False
                self.queue.all_tasks_done.wait(timeout=delay)

            return True
        finally:
            self.queue.all_tasks_done.release()

    def _thread_worker(self) -> None:
        failed_on_previous_iteration = False
        while True:
            item = self.queue.get()
            try:
                if item is _TERMINATOR:
                    break
                if not failed_on_previous_iteration:
                    self.item_processing_handler(item)
                else:
                    self.skip_processing_callback(item)
            except Exception as e:
                # It's better to continue cleanning the queue with queue.task_done() to avoid hanging on queue.join()
                failed_on_previous_iteration = True
                err_msg = "TrackingAsyncWebChannelBackgroundWorker will skip processing next events"
                log_exception(err_msg, e, logger)
            finally:
                self.queue.task_done()
            sleep(0)

    def _ensure_thread(self) -> None:
        if not self.is_alive:
            self.start()

    def submit(self, item: Any) -> None:
        self._ensure_thread()
        self.queue.put(item)

    def get_qsize(self) -> int:
        return self._queue.qsize() + 1


@attr.s
class AsyncWebChannelQueueItem(object):
    name: str = attr.ib()
    data: Dict[str, Any] = attr.ib()
    is_orchestration_run: str = attr.ib()
    stop_tracking_on_failure: str = attr.ib()


class TrackingAsyncWebChannel(TrackingChannel):
    """Json API client implementation with non-blocking calls."""

    def __init__(
        self,
        max_retries,
        remove_failed_store,
        tracker_raise_on_error,
        is_verbose,
        databand_api_client,
        *args,
        **kwargs,
    ):
        super(TrackingAsyncWebChannel, self).__init__(*args, **kwargs)

        self.web_channel = TrackingWebChannel(databand_api_client)

        self._background_worker = TrackingAsyncWebChannelBackgroundWorker(
            item_processing_handler=self._background_worker_item_handler,
            skip_processing_callback=self._background_worker_skip_processing_callback,
        )

        self._max_retries = max_retries
        self._remove_failed_store = remove_failed_store
        self._tracker_raise_on_error = tracker_raise_on_error
        self._is_verbose = is_verbose
        self._log_fn = logger.info if self._is_verbose else logger.debug
        self._shutting_down = False
        self._start_time = utcnow()

    def _handle(self, name, data):
        if self._shutting_down:
            # May happen if the store is used during databand ctx exiting
            raise RuntimeError("TrackingAsyncWebChannel is invoked during flush")

        # read tracking args from current runtime to avoid thread from reading a newer runtime context
        stop_tracking_on_failure = self._remove_failed_store or (
            in_tracking_run() and is_state_call(name)
        )

        # send data for processing in a thread
        item = AsyncWebChannelQueueItem(
            name=name,
            data=data,
            is_orchestration_run=is_orchestration_run(),
            stop_tracking_on_failure=stop_tracking_on_failure,
        )
        self._background_worker.submit(item=item)

    def _background_worker_item_handler(self, item: AsyncWebChannelQueueItem):
        try:
            tries = self._max_retries if is_state_call(item.name) else 1
            self._log_fn(
                "TrackingAsyncWebChannel.thread_worker processing %s", item.name
            )
            try_run_handler(tries, self.web_channel, item.name, {"data": item.data})
            self._log_fn(
                "TrackingAsyncWebChannel.thread_worker completed %s", item.name
            )

        except Exception as exc:
            log_exception_to_server()

            logger.warning(
                "Exception in TrackingAsyncWebChannel in background worker",
                exc_info=True,
            )
            if item.stop_tracking_on_failure:
                raise exc

            if isinstance(exc, DatabandWebserverNotReachableError):
                if in_tracking_run():
                    logger.warning(str(exc))

                if item.is_orchestration_run:
                    # in orchestration runs we have good error collection that's show error banner
                    # error should have good msg and no need to show full trace
                    exc.show_exc_info = False

                if self._tracker_raise_on_error:
                    raise exc
            # in all other cases continue

    def _background_worker_skip_processing_callback(
        self, item: AsyncWebChannelQueueItem
    ):
        self._log_fn(
            "TrackingAsyncWebChannel skips %s tracking event due to a previous failure",
            item.name,
        )

    def flush(self):
        # skip the handler if worker already exited to avoid hanging
        if not self._background_worker.is_alive:
            return
        # process remaining items in the queue

        tracking_duration = (utcnow() - self._start_time).in_seconds()
        # don't exceed 10% of whole tracking duration while flushing but not less then 5m and no more then 30m
        flush_limit = min(max(tracking_duration * 0.1, 5 * 60), 30 * 60)

        logger.info(
            "Waiting %ss for TrackingAsyncWebChannel to complete async tasks...",
            flush_limit,
        )
        self._shutting_down = True
        try:
            self._background_worker.flush(flush_limit)
            self.web_channel.flush()
            logger.info("TrackingAsyncWebChannel completed all tasks")
        except TimeoutError as e:
            err_msg = f"TrackingAsyncWebChannel flush exceeded {flush_limit}s timeout"
            log_exception(err_msg, e, logger)
        finally:
            self._shutting_down = False

    def is_ready(self):
        return self.web_channel.is_ready() and not self._shutting_down

    def __str__(self):
        return "AsyncWeb"



r"""Command-line tool to validate and pretty-print JSON

Usage::

    $ echo '{"json":"obj"}' | hjson
    {
        "json": "obj"
    }

"""
from __future__ import with_statement
import sys
import hjson
import pkg_resources  # part of setuptools

HELP="""Hjson, a user interface for JSON

Usage:
  hjson [options]
  hjson [options] <input>
  hjson (-h | --help)
  hjson (-V | --version)

Options:
  -h --help     Show this screen.
  -j            Output as formatted JSON.
  -c            Output as JSON.
  -V --version  Show version.
""";

def showerr(msg):
    sys.stderr.write(msg)
    sys.stderr.write("\n")

def main():
    format = 'hjson'
    args = []
    for arg in sys.argv[1:]:
        if arg == '-h' or arg == '--help':
            showerr(HELP)
            return
        elif arg == '-j': format = 'json'
        elif arg == '-c': format = 'compact'
        elif arg == '-V' or arg == '--version':
            showerr('Hjson ' + pkg_resources.require("Hjson")[0].version)
            return

        elif arg[0] == '-':
            showerr(HELP)
            raise SystemExit('unknown option ' + arg)
        else:
            args.append(arg)

    outfile = sys.stdout
    if len(args) == 0:
        infile = sys.stdin
    elif len(args) == 1:
        infile = open(args[0], 'r')
    else:
        showerr(HELP)
        raise SystemExit('unknown options')

    with infile:
        try:
            obj = hjson.load(infile, use_decimal=True)
        except ValueError:
            raise SystemExit(sys.exc_info()[1])

    with outfile:
        if format == 'json':
            hjson.dumpJSON(obj, outfile, ensure_ascii=False, use_decimal=True, indent='  ')
        elif format == 'compact':
            hjson.dumpJSON(obj, outfile, ensure_ascii=False, use_decimal=True, separators=(',', ':'))
        else:
            hjson.dump(obj, outfile, ensure_ascii=False, use_decimal=True)

        outfile.write('\n')

if __name__ == '__main__':
    main()



# © Copyright Databand.ai, an IBM Company 2022


class MultiTargetOpen(object):
    """FileInput([files[, inplace[, backup[, bufsize[, mode[, openhook]]]]]])

    Class FileInput is the implementation of the module; its methods
    filename(), lineno(), fileline(), isfirstline(), isstdin(), fileno(),
    nextfile() and close() correspond to the functions of the same name
    in the module.
    In addition it has a readline() method which returns the next
    input line, and a __getitem__() method which implements the
    sequence behavior. The sequence must be accessed in strictly
    sequential order; random access and readline() cannot be mixed.
    """

    def __init__(self, targets=None, mode="r"):

        self._targets = targets

        self._filename = None

        self._startlineno = 0
        self._filelineno = 0

        self._file_idx = 0
        self._file = None
        # restrict mode argument to reading modes
        if mode not in ("r", "rU", "U", "rb"):
            raise ValueError(
                "FileInput opening mode must be one of " "'r', 'rU', 'U' and 'rb'"
            )
        self._mode = mode

    def __del__(self):
        self.close()

    def close(self):
        try:
            self.nextfile()
        finally:
            self._file_idx = 0

    def __iter__(self):
        return self

    def __next__(self):
        while 1:
            line = self._readline()
            if line:
                self._filelineno += 1
                return line
            if not self._file:
                raise StopIteration
            self.nextfile()
            # repeat with next file

    def __getitem__(self, i):
        if i != self.lineno():
            raise RuntimeError("accessing lines out of order")
        try:
            return self.next()
        except StopIteration:
            raise IndexError("end of input reached")

    def nextfile(self):
        """
        close current file, that all
        """
        file = self._file
        self._file = None
        try:
            del self._readline  # restore FileInput._readline
        except AttributeError:
            pass
        if file:
            file.close()

    def readline(self):
        while 1:
            line = self._readline()
            if line:
                self._filelineno += 1
                return line
            if not self._file:
                return line
            self.nextfile()
            # repeat with next file

    def _readline(self):  # pylint: disable=method-hidden
        """
        first iteration runs here
        """
        if self._file_idx >= len(self._targets):
            return None

        self._filename = self._targets[self._file_idx]
        self._file_idx += 1

        self._startlineno = self.lineno()
        self._filelineno = 0

        # This may raise IOError
        self._file = self._filename.open(self._mode)

        self._readline = self._file.readline  # hide FileInput._readline
        return self._readline()

    def filename(self):
        return self._filename

    def lineno(self):
        return self._startlineno + self._filelineno

    def filelineno(self):
        return self._filelineno

    def fileno(self):
        if self._file:
            try:
                return self._file.fileno()
            except ValueError:
                return -1
        else:
            return -1

    def isfirstline(self):
        return self._filelineno == 1

    def read(self):
        raise NotImplementedError(
            "We don't support reading 'blob' from multiple files, use readlines function"
        )

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def readlines(self):
        return list(self)



# © Copyright Databand.ai, an IBM Company 2022

import enum


class ParameterScope(enum.Enum):
    """ParameterScope."""

    task = "task"  # do not propagate value
    children = "children"  # propagate value to direct children



# © Copyright Databand.ai, an IBM Company 2022

import logging

import attr

from dbnd._core.configuration.config_path import ConfigPath
from dbnd._core.constants import OutputMode, TaskEssence
from dbnd._core.errors import friendly_error
from dbnd._core.errors.friendly_error.task_parameters import (
    failed_to_build_parameter,
    unknown_value_type_in_parameter,
    value_is_not_parameter_cls,
)
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    _add_description,
    _ParameterKind,
)
from dbnd._core.parameter.validators import ChoiceValidator, NumericalValidator
from dbnd._core.utils.basics.nothing import NOTHING, is_defined, is_not_defined
from targets.target_config import TargetConfig
from targets.types import DataList
from targets.value_meta import ValueMetaConf
from targets.values import (
    TargetPathValueType,
    TargetValueType,
    ValueType,
    get_value_type_of_obj,
    get_value_type_of_type,
)
from targets.values.builtins_values import EnumValueType
from targets.values.structure import _StructureValueType
from targets.values.value_type import InlineValueType


logger = logging.getLogger(__name__)


class ParameterFactory(object):
    def __init__(self, parameter=None):
        self.parameter = parameter or ParameterDefinition()

    def modify(self, **kwargs):
        if not kwargs:
            return self
        new_parameter = attr.evolve(self.parameter, **kwargs)
        return ParameterFactory(parameter=new_parameter)

    def value(self, default=NOTHING, **kwargs):
        """
        Constructs parameter with default=Value
        :param default: the default value for this parameter. This should match the type of the
                        Parameter, i.e. ``datetime.date`` for ``DateParameter`` or ``int`` for
                        ``IntParameter``. By default, no default is stored and
                        the value must be specified at runtime.
        :param bool significant: specify ``False`` if the parameter should not be treated as part of
                                 the unique identifier for a Task. An insignificant Parameter might
                                 also be used to specify a password or other sensitive information
                                 that should not be made public. Default:
                                 ``True``.
        :param str description: A human-readable string describing the purpose of this Parameter.
                                For command-line invocations, this will be used as the `help` string
                                shown to users. Default: ``None``.
        :param OutputMode output_mode: specify OutputMode.prod_immutable to convert this parameter
                                to production immutable output
        :return: ParameterFactory
        """
        return self.modify(default=default, **kwargs)

    def __call__(self, **kwargs):
        """
        Changes parameter behaviour
        :param default: the default value for this parameter. This should match the type of the
                        Parameter, i.e. ``datetime.date`` for ``DateParameter`` or ``int`` for
                        ``IntParameter``. By default, no default is stored and
                        the value must be specified at runtime.
        :param bool significant: specify ``False`` if the parameter should not be treated as part of
                                 the unique identifier for a Task. An insignificant Parameter might
                                 also be used to specify a password or other sensitive information
                                 that should not be made public. Default:
                                 ``True``.
        :param str description: A human-readable string describing the purpose of this Parameter.
                                For command-line invocations, this will be used as the `help` string
                                shown to users. Default: ``None``.
        :param OutputMode output_mode: specify OutputMode.prod_immutable to convert this parameter
                                to production immutable output
        :return: ParameterFactory
        """
        return self.modify(**kwargs)

    @property
    def none(self):
        return self.modify(default=None)

    ########
    # formats
    def target_config(self, target_config):
        return self.modify(target_config=target_config)

    @property
    def _target_config(self):
        return self.parameter.target_config or TargetConfig()

    @property
    def gz(self):
        return self.target_config(self._target_config.gzip)

    @property
    def gzip(self):
        return self.target_config(self._target_config.gzip)

    @property
    def zip(self):
        return self.target_config(self._target_config.zip)

    @property
    def parquet(self):
        return self.target_config(self._target_config.parquet)

    @property
    def json(self):
        return self.target_config(self._target_config.json)

    @property
    def hdf5(self):
        return self.target_config(self._target_config.hdf5)

    @property
    def txt(self):
        return self.target_config(self._target_config.txt)

    @property
    def csv(self):
        return self.target_config(self._target_config.csv)

    @property
    def tsv(self):
        return self.target_config(self._target_config.tsv)

    @property
    def pickle(self):
        return self.target_config(self._target_config.pickle)

    @property
    def tfmodel(self):
        return self.target_config(self._target_config.tfmodel)

    @property
    def tfhistory(self):
        return self.target_config(self._target_config.tfhistory)

    @property
    def feather(self):
        return self.target_config(self._target_config.feather)

    @property
    def excel(self):
        return self.target_config(self._target_config.excel)

    def with_format(self, format):
        return self.target_config(self._target_config.with_format(format))

    @property
    def folder(self):
        """
        Input/Output path is a folder
        """
        return self.target_config(self._target_config.as_folder()).no_extension

    @property
    def folder_data(self):

        """
        Input/Output path is a folder with data (extension based on type)
        """
        return self.target_config(self._target_config.as_folder())

    def with_flag(self, flag=True):
        return self.target_config(self._target_config.with_flag(flag))

    @property
    def file_numpy(self):
        return self.target_config(self._target_config.numpy)

    @property
    def require_local_access(self):
        """
        Write parameter to local filesystem before syncing to remote filesystem
        :return: ParameterFactory
        """
        return self.target_config(self._target_config.with_require_local_access())

    @property
    def overwrite(self):
        return self.target_config(self._target_config.with_overwrite())

    def target_factory(self, target_factory, folder=False):
        tc = self.target_config(self._target_config.with_target_factory(target_factory))
        if folder:
            return tc.folder
        return tc

    def load_options(self, file_format, **options):
        current_options = self.parameter.load_options or {}
        current_options = current_options.copy()
        current_options[file_format] = options
        return self.modify(load_options=current_options)

    def save_options(self, file_format, **options):
        current_options = self.parameter.save_options or {}
        current_options = current_options.copy()
        current_options[file_format] = options
        return self.modify(save_options=current_options)

    def custom_target(self, target_factory, folder=False):
        return self.target_factory(target_factory, folder=folder)

    @property
    def no_extension(self):
        """
        output without extension
        :return: ParameterFactory
        """
        return self.modify(output_ext="")

    @property
    def no_env_interpolation(self):
        """
        Disables Environment Variables Interpolation in strings
        :return: ParameterFactory
        """
        return self.modify(env_interpolation=False)

    @property
    def disable_jinja_templating(self):
        """
        Disables Jinja template parsing
        :return: ParameterFactory
        """
        return self.modify(disable_jinja_templating=True)

    @property
    def prod_immutable(self):
        """
        Converts parameter to production immutable output
        :return: ParameterFactory
        """
        return self.modify(output_mode=OutputMode.prod_immutable)

    # kind changes
    @property
    def output(self):
        """
        Converts parameter to output
        :return: ParameterFactory
        """
        task_output = self.modify(kind=_ParameterKind.task_output)

        value_type = self.parameter.value_type
        if value_type is None:
            return task_output.target
        return task_output

    def cast(self, type_):
        return self

    def __getitem__(self, type_):
        # we want special treatment for Config classes
        if TaskEssence.CONFIG.is_instance(type_):
            return self.nested_config(type_)

        value_type = get_value_type_of_type(type_, inline_value_type=True)
        if not value_type:
            raise unknown_value_type_in_parameter(type_)
        return self.modify(value_type=value_type)

    def type(self, type_):
        value_type = get_value_type_of_type(type_)
        if not value_type:
            value_type = InlineValueType(type_=type_)
        return self.modify(value_type=value_type)

    def sub_type(self, sub_type):
        return self.modify(sub_type=sub_type)

    # behaviour modifiers
    @property
    def data(self):
        """
        mark output as data (it will not be loaded during the "build" stage, while we compile all tasks graph)
        """
        if self.parameter.value_type is None:
            return self.modify(value_type=TargetValueType, load_on_build=False)
        return self.modify(load_on_build=False)

    def help(self, help_message):
        """
        A human-readable string describing the purpose of this Parameter.
        For command-line invocations, this will be used as the `help` string shown to users. Default: ``None``
        :return: ParameterFactory
        """
        return self.modify(description=help_message)

    @property
    def non_significant(self):
        """Mark parameter to not be treated as part of the unique identifier for a Task.
        An insignificant Parameter might also be used to specify a password or other
        sensitive information that should not be made public.
        :return: ParameterFactory
        """
        return self.modify(significant=False)

    def config(self, section, name):
        return self.modify(config_path=ConfigPath(section=section, key=name))

    def description(self, help_message):
        return self.modify(description=help_message)

    def default(self, value):
        return self.modify(default=value)

    def enum(self, enum, **kwargs):
        """
        A parameter whose value is an :class:`~enum.Enum`.

        In the task definition, use

        .. code-block:: python

            class Model(enum.Enum):
              Honda = 1
              Volvo = 2

            class MyTask(dbnd.Task):
              my_param = databand.EnumParameter(enum=Model)

        At the command line, use,

        .. code-block:: console

            $ dbnd MyTask --my-param Honda

        """
        return self(**kwargs).type(EnumValueType(enum))

    def choices(self, choices):
        """
            Consider using :class:`parameter.enum` for a typed, structured
            alternative.  This parameter can perform the same role when all choices are the
            same type and transparency of parameter value on the command line is
            desired.
        :param choices: An iterable, all of whose elements are of `T` to
                        restrict parameter choices to.
        :return:
        """
        return self.modify(validator=ChoiceValidator(choices))

    def numerical(self, min_value, max_value):
        return self.modify(validator=NumericalValidator(min_value, max_value))

    def custom_datetime(self, interval=1, start=None):
        custom_value_type = self.parameter.value_type.__class__(interval, start)
        return self.type(custom_value_type)

    def nested_config(self, config_cls):
        from dbnd._core.parameter.value_types.task_value import ConfigValueType

        return self(
            value_type=ConfigValueType(config_cls),
            # default=config_cls,
            significant=False,  # doesn't affect output
        )

    @property
    def c(self):
        """
        config parameter
        """
        return self.modify(empty_default=True)

    @property
    def system(self):
        return self.modify(system=True, significant=False)

    # type aliases
    @property
    def target(self):
        return self.type(TargetValueType)

    @property
    def path_str(self):
        return self.type(TargetPathValueType)

    @property
    def pandas_dataframe(self):
        import pandas

        return self[pandas.DataFrame]

    @property
    def data_list_str(self):
        return self[DataList[str]]

    @property
    def data_list(self):
        return self[DataList]

    @property
    def numpy_array(self):
        import numpy

        return self[numpy.ndarray]

    @property
    def _p(self):
        """
        inline builder for tests purposes only!
        :return:
        """
        return self.build_parameter("inline")

    def build_parameter(self, context):
        try:
            return self._build_parameter(context)
        except Exception as ex:
            raise failed_to_build_parameter(context, self.parameter, ex)

    def _build_value_type(self, context):
        s = self.parameter
        value_type = s.value_type
        default = s.default
        if value_type is None:
            if is_defined(default):
                value_type = get_value_type_of_obj(default)
                if value_type is None:
                    raise friendly_error.task_parameters.no_value_type_from_default(
                        default, context=context
                    )
            elif value_type is None:  # we don't have value type! let's fail!
                if s.load_on_build is False:  # we are in data mode
                    s.value_type = TargetValueType
                else:
                    raise friendly_error.task_parameters.no_value_type_defined_in_parameter(
                        context=context
                    )
        else:
            # let validate that what we have can be ValueType!
            resolved_value_type = get_value_type_of_type(value_type)
            if resolved_value_type is None:  # we don't have value type! let's fail!
                raise friendly_error.task_parameters.unknown_value_type_in_parameter(
                    value_type
                )
            value_type = resolved_value_type  # type: ValueType

        if s.sub_type:
            sub_value_type = get_value_type_of_type(s.sub_type)
            if isinstance(value_type, _StructureValueType):
                value_type = value_type.__class__(sub_value_type=sub_value_type)
            else:
                raise friendly_error.task_parameters.sub_type_with_non_structural_value(
                    context=context, value_type=value_type, sub_type=s.sub_type
                )
        return value_type

    def _build_parameter(self, context="inline"):
        s = self.parameter  # type: ParameterDefinition
        update_kwargs = {}

        value_type = self._build_value_type(context)

        validator = s.validator
        if s.choices:
            validator = ChoiceValidator(s.choices)

        if is_not_defined(s.default):
            if s.empty_default:
                update_kwargs["default"] = value_type._generate_empty_default()

        if not is_defined(s.load_on_build):
            update_kwargs["load_on_build"] = value_type.load_on_build

        # create value meta
        if s.value_meta_conf is None:
            update_kwargs["value_meta_conf"] = ValueMetaConf(
                log_preview=s.log_preview,
                log_preview_size=s.log_preview_size,
                log_schema=s.log_schema,
                log_size=s.log_size,
                log_stats=s.log_stats,
                log_histograms=s.log_histograms,
            )

        # Whether different values for this parameter will differentiate otherwise equal tasks
        description = s.description or ""
        if not is_defined(description):
            if s.is_output() and s.default_output_description:
                description = s.default_output_description
            elif not s.load_on_build and s.default_input_description:
                description = s.default_input_description
            else:
                description = s.default_description

            if s.validator:
                description = _add_description(description, validator.description)
            update_kwargs["description"] = description()
        # We need to keep track of this to get the order right (see Task class)
        ParameterDefinition._total_counter += 1
        if s.kind == _ParameterKind.task_output:
            update_kwargs["significant"] = False

        updated = self.modify(
            value_type=value_type,
            value_type_defined=value_type,
            validator=validator,
            description=description,
            parameter_id=ParameterDefinition._total_counter,
            **update_kwargs
        )

        return updated.parameter


PARAMETER_FACTORY = ParameterFactory()
parameter = PARAMETER_FACTORY
parameter.__doc__ = """
    Makes a class variable become a Task or Pipeline parameter.

    Parameters are DBND's equivalent to creation of a constructor for each task.
    This is only used when using a Task or Pipeline class, otherwise the function's parameters are automatically made into
    DBND parameter.
    In this example, ``data`` is set as PrepareData's parameter::

        class PrepareData(Task):
            data = parameter.data
            ..."""

output = parameter.output  # type: ParameterFactory
output.__doc__ = """
    Makes a class variable become a Task or Pipeline Output.

    You can customize the output type, and it can be set inside Run or Band.
    This is only used when using a Task or Pipeline class, otherwise the function's return value is automatically set as
    the Task or Pipeline output.

    In this example, prepared_data is set as PrepareData's output, and is set inside the run method::

        class PrepareData(Task):
            data = parameter.data

            prepared_data = output.csv.data

            def run(self):
                self.prepared_data = do_something_with_data(self.data)"""
data = parameter.target.data  # type: ParameterFactory
data.__doc__ = """Marks a parameter as data."""


def build_parameter(parameter, context="inline"):
    try:
        if isinstance(parameter, ParameterFactory):
            return parameter.build_parameter(context)
        if isinstance(parameter, type) and issubclass(parameter, ParameterDefinition):
            # may be used just used name of the class
            return parameter()
    except Exception:
        logger.warning("Failed to build parameter for '%s'", context)
        raise
    if not isinstance(parameter, ParameterDefinition):
        raise value_is_not_parameter_cls(parameter, context=context)

    return parameter



# © Copyright Databand.ai, an IBM Company 2022

import logging

import six


logger = logging.getLogger(__file__)


def inverse_dict(x):
    """Make keys or values and values of keys"""
    return {v: k for k, v in six.iteritems(x)}


def list_to_tuple(x):
    """Make tuples out of lists and sets to allow hashing"""
    if isinstance(x, list) or isinstance(x, set):
        return tuple(x)
    else:
        return x


class MappingCombineStrategy(object):
    replace_and_warn = "replace_and_warn"
    replace = "replace"


def combine_mappings(left, right, strategy=MappingCombineStrategy.replace):
    left = left or {}
    right = right or {}

    # immidiate return if similar of empty
    if not left:
        return right.copy()

    result = left.copy()
    if left == right or not right:
        return result

    for key, value in six.iteritems(right):
        if strategy == MappingCombineStrategy.replace_and_warn and key in result:
            logger.warning("Overriding '%s' value with '%s'", key, value)

        result[key] = value
    return result


def split_list(func, iter):
    left, right = [], []
    for v in iter:
        if func(v):
            left.append(v)
        else:
            right.append(v)
    return left, right


def list_of_strings(value_list):
    if not value_list:
        return []
    return list(map(str, value_list))


def get_object_attrs(cls):
    params = []
    for param_name in dir(cls):
        param_obj = getattr(cls, param_name)
        params.append((param_name, param_obj))
    return params



# © Copyright Databand.ai, an IBM Company 2022

import inspect


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def _as_dotted_dict(**dict_obj):
    return dotdict(**dict_obj)


def build_dict_from_instance_properties(class_instance) -> dict:
    """Returns dict of all class properties including @property methods"""
    properties = {}
    for prop_name in dir(class_instance):
        prop_value = getattr(class_instance, prop_name)
        if not prop_name.startswith("__") and not inspect.ismethod(prop_value):
            properties[prop_name] = prop_value
    return properties


class rdotdict(dict):
    """
    this is "recursive" dotdict - provides dot-notation access to dict, and recursively
    adds itself for internal values (dicts and lists). So this code is valid:
    ```
    d = {"a": [{"b": 42}]}
    rdotdict(d).a[0].b == 42
    ```
    """

    def __getattr__(self, item):
        return self[item]

    def __getitem__(self, item):
        return rdotdict.try_create(dict.__getitem__(self, item))

    @staticmethod
    def try_create(val):
        """Applies rdotdict for rdotdict-able object (i.e dict or list)"""
        if isinstance(val, dict):
            return rdotdict(**val)
        if isinstance(val, list):
            return [rdotdict.try_create(x) for x in val]
        return val



# © Copyright Databand.ai, an IBM Company 2022

from collections import defaultdict
from itertools import chain
from typing import Dict, Iterable, List, Optional, Tuple
from urllib.parse import urlparse

import attr

from more_itertools import first, padded

from dbnd._core.constants import DbndDatasetOperationType, DbndTargetOperationType
from dbnd._core.utils.sql_tracker_common.sql_extract import Column, Schema
from dbnd._core.utils.sql_tracker_common.utils import strip_quotes
from dbnd.utils.anonymization import secrets_anonymizer
from targets.connections import build_conn_path
from targets.value_meta import ValueMeta
from targets.values import register_value_type
from targets.values.builtins_values import DataValueType
from targets.values.pandas_histograms import PandasHistograms


class Connection:
    @property
    def host(self) -> str:
        ...

    @property
    def port(self) -> Optional[int]:
        ...

    @property
    def database(self) -> Optional[str]:
        ...

    @property
    def schema(self) -> Optional[str]:
        ...


DTypes = Dict[str, str]


@attr.s
class SqlOperation:
    extracted_schema: Schema = attr.ib()
    dtypes: DTypes = attr.ib()
    records_count: int = attr.ib()
    query: str = attr.ib()
    query_id: str = attr.ib()
    success: bool = attr.ib()
    op_type: DbndDatasetOperationType = attr.ib()
    error: str = attr.ib()
    dataframe = attr.ib(default=None)  # type: pd.DataFrame

    def __attrs_post_init__(self):
        self.query = secrets_anonymizer.anonymize(self.query)

    @property
    def columns(self) -> List[str]:
        try:
            return list(self.dtypes)
        except TypeError:
            return None

    @property
    def columns_count(self) -> int:
        try:
            return len(self.dtypes)
        except TypeError:
            return None

    @property
    def is_file(self) -> bool:
        # files are in scope of read operations
        return (
            self.op_type is DbndTargetOperationType.read
            and first(self.extracted_columns).is_file
        )

    @property
    def is_stage(self) -> bool:
        # stage are in scope of read operations
        return (
            self.op_type is DbndTargetOperationType.read
            and first(self.extracted_columns).is_stage
        )

    @property
    def tables(self) -> Iterable[str]:
        return [col.dataset_name for col in self.extracted_columns]

    @property
    def extracted_columns(self) -> Iterable[Column]:
        return chain.from_iterable(self.extracted_schema.values())

    def evolve_schema(
        self, tables_schemas: Dict[str, DTypes], file_schema: DTypes = None
    ) -> "SqlOperation":
        if self.dtypes is not None:
            return self

        dtypes = self.build_dtypes(tables_schemas, file_schema)
        return attr.evolve(self, dtypes=dtypes)

    def build_dtypes(
        self, tables_schemas: Dict[str, DTypes], file_schema: DTypes = None
    ) -> DTypes:
        dtypes = {}
        for col in self.extracted_columns:
            if col.is_file or col.is_stage:
                if file_schema:
                    dtypes.update(file_schema)
                else:
                    continue
            if col.is_wildcard:
                all_columns = tables_schemas.get(col.dataset_name, {})
                dtypes.update(all_columns)
            else:
                col_name = col.name.strip('"')
                col_type = tables_schemas.get(col.dataset_name, {}).get(
                    col_name.lower()
                )
                if col_type:
                    dtypes[col_name] = col_type
        return dtypes if dtypes else None

    def evolve_table_name(self, connection: Connection) -> "SqlOperation":
        schema: Schema = defaultdict(list)
        for name, cols in self.extracted_schema.items():
            for col in cols:
                if not col.is_file:
                    table_name = render_table_name(connection, col.dataset_name)
                else:
                    table_name = col.dataset_name
                col = attr.evolve(col, dataset_name=table_name)
                schema[name].append(col)

        return attr.evolve(self, extracted_schema=schema)


def split_table_name(table_name: str) -> Tuple[str, Optional[str], Optional[str]]:
    table, schema, database = padded(reversed(table_name.split(".")[:3]), None, 3)
    return database, schema, table


def render_table_name(connection: Connection, table_name: str, sep: str = ".") -> str:
    database, schema, table = split_table_name(table_name)
    database = connection.database or database
    schema = connection.schema or schema
    path = sep.join(map(strip_quotes, filter(None, [database, schema, table])))
    return path


def render_connection_path(
    connection: Connection, operation: SqlOperation, conn_type: str
) -> str:
    # We takes only the first table as a workaround
    table_name = first(operation.tables)
    if operation.is_file:
        urlparsed = urlparse(table_name)
        return build_conn_path(
            conn_type=urlparsed.scheme,
            hostname=urlparsed.netloc,
            port=urlparsed.port,
            path=urlparsed.path,
        )
    else:
        path = render_table_name(connection, table_name, sep="/")
    return build_conn_path(
        conn_type=conn_type, hostname=connection.host, port=connection.port, path=path
    )


class SqlOperationValueMeta(DataValueType):
    type = SqlOperation
    type_str = "SqlOperation"
    support_merge = False

    config_name = "sql_operation"
    is_lazy_evaluated = False

    def get_value_meta(self, value: SqlOperation, meta_conf):
        data_schema = {}
        data_dimensions = None

        if meta_conf.log_schema:
            data_schema = {"type": self.type_str, "dtypes": value.dtypes}

        if meta_conf.log_size:
            data_dimensions = [value.records_count, value.columns_count]
            data_schema["shape"] = data_dimensions
            # todo: size?

        # currently columns_stats and histogram are not supported
        columns_stats, histograms = [], {}
        hist_sys_metrics = None
        if meta_conf.log_stats and value.dataframe is not None:
            columns_stats, _ = PandasHistograms(
                value.dataframe, meta_conf
            ).get_histograms_and_stats()

        return ValueMeta(
            value_preview=None,
            data_dimensions=data_dimensions,
            query=value.query,
            data_schema=data_schema,
            data_hash=str(hash(self.to_signature(value))),
            columns_stats=columns_stats,
            histogram_system_metrics=hist_sys_metrics,
            histograms=histograms,
        )


register_value_type(SqlOperationValueMeta())



# © Copyright Databand.ai, an IBM Company 2022


def _task_name(task):
    return task.task_definition.full_task_family


def _band_call_str(task):
    if not task:
        return
    if task.task_decorator:
        return "%s()" % _task_name(task)
    return "%s.band()" % _task_name(task)


def _run_name(task):
    return task.task_definition.run_name()


def _safe_task_family(task):
    if not task:
        return "unknown"

    from dbnd._core.task import Task

    if isinstance(task, Task):
        return _task_name(task)
    return str(task)


def _parameter_name(task, parameter):
    return "%s.%s" % (task.friendly_task_name, parameter.name)


def _safe_target(target):
    msg = str(target)
    if target and target.source_task:
        msg += " created by {task}".format(task=_safe_task_family(target.source_task))
    return msg



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors.base import (
    DatabandBuildError,
    DatabandConfigError,
    DatabandError,
    DatabandRunError,
    DatabandRuntimeError,
    DatabandSystemError,
    MissingParameterError,
    ParameterError,
    ParseParameterError,
    ParseValueError,
    TaskClassAmbigiousException,
    TaskClassNotFoundException,
    TaskValidationError,
    UnknownParameterError,
    ValueTypeError,
)
from dbnd._core.errors.errors_utils import get_help_msg, show_exc_info


__all__ = [
    "DatabandBuildError",
    "DatabandConfigError",
    "DatabandError",
    "DatabandRunError",
    "DatabandRuntimeError",
    "DatabandSystemError",
    "MissingParameterError",
    "ParameterError",
    "ParseParameterError",
    "ParseValueError",
    "TaskClassAmbigiousException",
    "TaskClassNotFoundException",
    "TaskValidationError",
    "UnknownParameterError",
    "ValueTypeError",
    "get_help_msg",
    "show_exc_info",
]



# © Copyright Databand.ai, an IBM Company 2022

# patch windows module so python-daemon will work on windows
# this is required for python-daemon<2.2 imported by airflow 1.10.1.
# This issue is resolved in newer versions of python-daemon


def getpwuid(*args, **kwargs):
    return -1



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.log.dbnd_log import (
    dbnd_log_debug,
    dbnd_log_exception,
    dbnd_log_info_error,
    dbnd_log_init_msg,
    dbnd_log_tracking,
    is_verbose,
)


__all__ = [
    "dbnd_log_debug",
    "dbnd_log_info_error",
    "is_verbose",
    "dbnd_log_exception",
    "dbnd_log_tracking",
    "dbnd_log_init_msg",
]



# © Copyright Databand.ai, an IBM Company 2022

import re

from typing import List


DEFAULT_MAKING_VALUE = "***"


class DataAnonymizer:
    def __init__(self, pattern: str, masking_expression: str = DEFAULT_MAKING_VALUE):
        self.pattern = pattern
        self.masking_expression = masking_expression

    def anonymize(self, raw_data):
        if not raw_data:
            return raw_data

        anonymize_data = raw_data
        finds = re.findall(self.pattern, anonymize_data, flags=re.IGNORECASE)
        for find in finds:
            if isinstance(find, tuple):
                for item in find:
                    anonymize_data = anonymize_data.replace(
                        item, self.masking_expression
                    )
            else:
                anonymize_data = anonymize_data.replace(find, self.masking_expression)

        return anonymize_data


class CompositeDataAnonymizer:
    def __init__(self, data_anonymizers: List[DataAnonymizer]):
        self.data_anonymizers = data_anonymizers

    @classmethod
    def from_string_patterns(cls, patterns: List[str]):
        anonymizers = [DataAnonymizer(pattern) for pattern in patterns]
        return cls(data_anonymizers=anonymizers)

    def anonymize(self, data: str):
        res = data
        for anonymizer in self.data_anonymizers:
            res = anonymizer.anonymize(res)

        return res


secrets_regexp = [
    r"::(?P<IAM_TO_KEEP>\d{12}):role",  # IAM Role
    r"aws(?:_access)?_key_id[\s=']*(AKI[a-zA-Z0-9]+)[';]",  # Redshift access key
    r"aws_secret(?:_access)?_key[\s=']*([a-zA-Z0-9/]+)[';]",  # Redshift secret key
    r"DATABAND_ACCESS_TOKEN=([a-zA-Z0-9\.]+)",  # Databand access token
]

secrets_anonymizer = CompositeDataAnonymizer.from_string_patterns(secrets_regexp)



# © Copyright Databand.ai, an IBM Company 2022

from typing import Any

import attr

from dbnd._core.errors import friendly_error
from dbnd._core.errors.friendly_error.task_execution import (
    failed_to_read_value_from_target,
)
from dbnd._core.parameter.parameter_definition import ParameterDefinition, T
from targets.errors import NotSupportedValue
from targets.multi_target import MultiTarget


class ResultProxyTarget(MultiTarget):
    target_no_traverse = True

    def __init__(self, source, names, properties=None):
        super(ResultProxyTarget, self).__init__(
            targets=None, properties=properties, source=source
        )
        self.names = names

    @property
    def targets(self):
        return list(self)

    def get_sub_result(self, name):
        return getattr(self.source_task, name)

    def __iter__(self):
        for name in self.names:
            yield self.get_sub_result(name)

    def __repr__(self):
        return "result(%s)" % ",".join(self.names)

    def __getitem__(self, item):
        if isinstance(item, int):  # handle p[0]
            try:
                item = self.names[item]
            except IndexError:
                raise IndexError(
                    "ResultProxyTarget index out fo the range. This target has only {} elements".format(
                        len(self.names)
                    )
                )
        try:
            return self.get_sub_result(item)
        except AttributeError:
            raise AttributeError("ResultProxyTarget has no `{}` target".format(item))

    def as_dict(self):
        return {name: self.get_sub_result(name) for name in self.names}

    def as_task(self):
        return self.source.task


@attr.s(hash=False, repr=False, str=False)
class FuncResultParameter(ParameterDefinition):
    schema = attr.ib(default=None)

    @property
    def names(self):
        return [p.name for p in self.schema]

    def build_output(self, task):
        return ResultProxyTarget(source=self._target_source(task), names=self.names)

    def dump_to_target(self, target, value, **kwargs):
        """
        We don't need to dump this value, it's just a map to all other outputs
        """

    def _validate_result(self, result):
        if not isinstance(result, (tuple, list, dict)):
            raise friendly_error.task_execution.wrong_return_value_type(
                self.task_definition, self.names, result
            )
        elif len(result) != len(self.schema):
            raise friendly_error.task_execution.wrong_return_value_len(
                self.task_definition, self.names, result
            )
        if isinstance(result, dict):
            if set(result.keys()).symmetric_difference(set(self.names)):
                raise NotSupportedValue(
                    "Returned result doesn't match expected schema. Expected {}, got {}".format(
                        self.names, result.keys()
                    )
                )

    def named_results(self, result):
        self._validate_result(result)
        if isinstance(result, dict):
            return [(name, result[name]) for name in self.names]
        return zip(self.names, result)

    def load_from_target(
        self, target, **kwargs
    ):  # type: (FuncResultParameter, ResultProxyTarget, **Any)-> T
        results = []
        for p in self.schema:
            p_target = target.get_sub_result(p.name)
            try:

                results.append(p.load_from_target(p_target, **kwargs))
            except Exception as ex:
                raise failed_to_read_value_from_target(
                    ex, p_target.source.task, p, target
                )

        return tuple(results)



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import absolute_import

import difflib

from dbnd._core.errors import DatabandBuildError, UnknownParameterError
from dbnd._core.errors.friendly_error.helpers import _band_call_str


def no_databand_context():
    return DatabandBuildError(
        "You are trying to create task without having active Databand context! "
        "You should have databand context while building/running tasks. "
        "You can create one inplace by adding `init_dbnd()` call"
    )


def unknown_parameter_in_constructor(constructor, param_name, task_parent):
    help_msg = "Remove {param_name} from constructor {constructor}".format(
        param_name=param_name, constructor=constructor
    )

    if task_parent:
        help_msg += " at %s method!" % _band_call_str(task_parent)

    return UnknownParameterError(
        "Unknown parameter '{param_name}' at {constructor}".format(
            constructor=constructor, param_name=param_name
        ),
        help_msg=help_msg,
    )


def unknown_parameter_in_config(
    task_name, param_name, source, task_param_names, config_type
):
    close_matches = difflib.get_close_matches(param_name, task_param_names)
    did_you_mean = None
    if close_matches:
        did_you_mean = "Did you mean: %s" % (", ".join(close_matches))
        help_msg = did_you_mean
    else:
        help_msg = "Remove '{param_name}' from the configuration.".format(
            param_name=param_name
        )
    help_msg = (
        help_msg
        + "\nAlternatively you can disabled this validation or set it to warning: validate_no_extra_params = warn/disabled under [{config_type}] in your config".format(
            config_type=config_type
        )
    )

    e = UnknownParameterError(
        "Unknown parameter '{param_name}' (source: {source}) in config for {task_or_section}".format(
            task_name=task_name,
            param_name=param_name,
            source=source,
            task_or_section="section [%s]" % task_name
            if config_type == "config"
            else "task '%s'" % task_name,
        ),
        help_msg=help_msg,
    )
    e.did_you_mean = did_you_mean
    return e


def pipeline_task_has_unassigned_outputs(task, param):
    return DatabandBuildError(
        "You have unassigned output '{param.name}' in task '{task}'".format(
            param=param, task=task
        ),
        help_msg="Check your {band} logic, Add self.{param.name} = SOME_TASK_OUTPUT".format(
            band=_band_call_str(task), param=param
        ),
    )


def failed_to_call_band(ex, task):
    return DatabandBuildError(
        "Failed to call '%s': %s" % (_band_call_str(task), ex),
        show_exc_info=False,
        nested_exceptions=[ex],
        help_msg="Check your %s logic" % _band_call_str(task),
    )


def failed_to_call(ex, task_cls):
    return DatabandBuildError(
        "Failed to invoke '%s': %s" % (_band_call_str(task_cls), ex),
        show_exc_info=False,
        nested_exceptions=[ex],
        help_msg="Check your %s logic" % _band_call_str(task_cls),
    )


def failed_to_assign_param_value_at_band(ex, param, value, task):
    return DatabandBuildError(
        "Failed to assign '{value}' to parameter {param.name} at '{band}': {ex}".format(
            band=_band_call_str(task), value=value, ex=ex, param=param
        ),
        show_exc_info=False,
        nested_exceptions=[ex],
        help_msg="Check your %s logic" % _band_call_str(task),
    )


def failed_to_build_output_target(p_name, task, ex):
    help_msg = (
        "Make sure that output parameter '{p_name}' is "
        "configured without keyword [parameter]!"
    ).format(p_name=p_name)
    return DatabandBuildError(
        (
            "Failed to build output target for parameter '{p_name}' in task '{task}'! "
            "Exception: {ex}"
        ).format(p_name=p_name, task=task, ex=ex),
        help_msg=help_msg,
    )


def iteration_over_task(task):
    help_msg = "You can iterate over task results, but not the task itself"
    if hasattr(task, "result"):
        help_msg = (
            "If you want to access task results, use {task}().result notation."
            " Probably this task has outputs defined via return value as well as regular parameter".format(
                task=task.get_task_family()
            )
        )
    # we have to return type error,
    # as sometimes we traverse task..and we need indication that it can't be  traversed
    # we can't use BuildError here
    return TypeError(
        "Task '{task}' can not be unpacked or iterated. {help_msg}".format(
            task=task.get_task_family(), help_msg=help_msg
        )
    )


def failed_to_import_pyspark(task, ex):
    return DatabandBuildError(
        "Tried to create spark session for a task {task} but failed to import pyspark".format(
            task=task.get_task_family
        ),
        show_exc_info=False,
        nested_exceptions=[ex],
        help_msg="Check your environment for pyspark installation.",
    )


def failed_to_access_dbnd_home(dbnd_home, ex):
    return DatabandBuildError(
        (
            "Failed to access DBND_HOME '{dbnd_home}'. "
            "Check that folder exists and a process has sufficient permissions to write there. "
            "Exception: {ex}"
        ).format(dbnd_home=dbnd_home, ex=ex)
    )


def incomplete_output_found_for_task(task_name, complete_outputs, incomplete_outputs):
    return DatabandBuildError(
        "Task {} has incomplete outputs! "
        "This means the task might fail every time. "
        "Complete outputs: {} "
        "Incomplete outputs: {} "
        "Hint: clean the environment or overwrite the output. "
        "To ignore this error, turn off 'validate_task_outputs_on_build' in the '[run]' configuration section".format(
            task_name,
            ", ".join(map(str, complete_outputs)),
            ", ".join(map(str, incomplete_outputs)),
        )
    )



# © Copyright Databand.ai, an IBM Company 2022

import typing

from typing import Type

import attr

from dbnd._core.task_build.task_const import _SAME_AS_PYTHON_MODULE
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._core.utils.basics.nothing import NOTHING, is_defined


if typing.TYPE_CHECKING:
    from dbnd._core.task import Task


def _short_name(name):
    # type: (str)->str
    """
    from my_package.sub_package  -> m.s
    """
    return ".".join(n[0] if n else n for n in name.split("."))


def format_source_suffix(name):
    return "[%s]" % name


@attr.s
class TaskPassport(object):
    """
    Task passport represents "name" identifiers for the Task/func/callable
    wrapped for Orchestration/Tracking by dbnd
    """

    full_task_family = attr.ib()
    full_task_family_short = attr.ib()
    task_family = attr.ib()
    task_config_section = attr.ib()

    @classmethod
    def from_task_cls(cls, task_class):
        # type: (Type[Task]) -> TaskPassport
        if task_class.task_decorator:
            return task_class.task_decorator.task_passport

        cls_name = task_class.__name__
        task_namespace = task_class.task_namespace
        module_name = task_class.__module__

        return cls.build_task_passport(
            cls_name=cls_name,
            module_name=module_name,
            task_namespace=task_namespace,
            task_family=task_class._conf__task_family,
        )

    @classmethod
    def from_module(cls, module):
        full_task_family_short = "%s" % (_short_name(module))

        return TaskPassport(
            full_task_family=module,
            full_task_family_short=full_task_family_short,
            task_family=module,
            task_config_section=None,
        )

    @classmethod
    def build_task_passport(
        cls, cls_name, module_name, task_namespace=NOTHING, task_family=None
    ):
        full_task_family = "%s.%s" % (module_name, cls_name)
        full_task_family_short = "%s.%s" % (_short_name(module_name), cls_name)

        if not is_defined(task_namespace):
            namespace_at_class_time = get_task_registry().get_namespace(module_name)
            if namespace_at_class_time == _SAME_AS_PYTHON_MODULE:
                task_namespace = module_name
            else:
                task_namespace = namespace_at_class_time

        if task_family:
            task_config_section = task_family
        elif task_namespace:
            task_family = "{}.{}".format(task_namespace, cls_name)
            task_config_section = task_family
        else:
            task_family = cls_name
            task_config_section = full_task_family

        return TaskPassport(
            full_task_family=full_task_family,
            full_task_family_short=full_task_family_short,
            task_family=task_family,
            task_config_section=task_config_section,
        )

    def format_source_name(self, name):
        return "%s%s" % (self.full_task_family_short, format_source_suffix(name))



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel


class TrackingWebChannel(TrackingChannel):
    """Json API client implementation."""

    def __init__(self, databand_api_client, *args, **kwargs):
        super(TrackingWebChannel, self).__init__(*args, **kwargs)
        self.client = databand_api_client

    def _handle(self, name, data):
        # this method may raise an  exception  in case of connectivity/server errors
        # such exceptions will be handled on upper layer
        return self.client.api_request("tracking/%s" % name, data)

    def is_ready(self):
        return self.client.is_ready()

    def __str__(self):
        return "Web"



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._vendor.marshmallow import fields, validate


class JobSchemaV2(ApiStrictSchema):
    id = fields.Int()
    name = fields.Str()
    user = fields.Str()
    reported_by_user = fields.Str()
    is_archived = fields.Boolean()
    ui_hidden = fields.Boolean()
    is_airflow_synced = fields.Boolean()
    project_id = fields.Int()
    project_name = fields.Str()

    # computed
    run_states = fields.Dict()
    source_link = fields.Str()

    # joined
    latest_run_start_time = fields.DateTime()
    latest_run_state = fields.Str()
    latest_run_uid = fields.UUID()
    latest_run_root_task_run_uid = fields.UUID()
    latest_run_trigger = fields.Str()
    latest_run_env = fields.Str()
    source_instance_name = fields.Str()
    source_type = fields.Str()
    airflow_instance_name = fields.Str()  # TODO_SERVER: API: Deprecate


class JobsSetArchiveSchema(ApiStrictSchema):
    ids = fields.List(fields.Integer(), required=True, validate=validate.Length(min=1))
    is_archived = fields.Boolean(required=True)


jobs_set_archive_schema = JobsSetArchiveSchema()



# © Copyright Databand.ai, an IBM Company 2022

CREATE_HELP_MSG = """
Create alert definition for given pipeline\n
\n
EXAMPLES\n
        dbnd alerts create --pipeline my_pipeline --severity MEDIUM run-duration --op '==' -v 100\n
        dbnd alerts create --pipeline my_pipeline --severity HIGH run-duration --op ML --look-back 10 --sensitivity 4\n
        dbnd alerts create --pipeline my_pipeline --severity HIGH run-duration --op range --baseline 10 --range 4\n
        dbnd alerts create --pipeline my_pipeline --severity MEDIUM run-state -v failed\n
        dbnd alerts create --pipeline my_pipeline --severity CRITICAL ran-last-x-seconds -v 120\n
        dbnd alerts create --pipeline my_pipeline --severity MEDIUM --task my_pipeline_task task-state -v cancelled\n
        dbnd alerts create --pipeline my_pipeline --severity CRITICAL --task my_pipeline_task custom-metric --metric-name my_metric --op '<=' -v 100 --str-value false\n
"""

LIST_HELP_MSG = """
List alert definitions for a given pipeline\n
\n
EXAMPLE\n
        dbnd alerts list --pipeline my_pipeline
"""

DELETE_HELP_MSG = """
Delete alert definitions using one of [uid / pipeline / wipe (all)]\n
\n
EXAMPLES\n
        dbnd alerts delete --uid my_alert_uid\n
        dbnd alerts delete --pipeline my_pipeline\n
        dbnd alerts delete --wipe\n
"""

RANGE_ALERT_BASE_LINE_HELP = "Base number from which range alert will be measured"
RANGE_ALERT_RANGE_HELP = "Percent of baseline to include in range alerts."
ANOMALY_ALERT_LOOK_BACK_HELP = "Look Back number of runs used to train anomaly alerts."
ANOMALY_ALERT_SENSITIVITY_HELP = "Sensitivity of anomaly detection alerts."
METRIC_HELP = "Custom metric name to alert"
IS_STR_VALUE_HELP = (
    "Custom metric alert value type (true for string value, false for float)"
)



# © Copyright Databand.ai, an IBM Company 2022

import inspect
import re

from typing import Callable, Dict, Iterable, Optional


TYPE_HINT_RE = re.compile(r"#type:\((.*)\)->(.+)")


class InvalidTypeHint(Exception):
    pass


# re:   Tuple[ Type1, Type2]
_TUPLE_TYPE_RE = re.compile(r"^Tuple\[(.+)]$")


def get_Tuple_params(type_str):
    tuple_def = _TUPLE_TYPE_RE.search(type_str)
    if tuple_def:
        return _split(tuple_def.group(1))
    return None


# re:   ( Type1, Type2,... )
_LIST_TYPE_RE = re.compile(r"^\((.+)\)$")


def get_inline_Tuple_params(type_str):
    list_type = _LIST_TYPE_RE.search(type_str)
    if list_type:
        return _split(list_type.group(1))
    return None


_SPLIT_ARGS_RE = re.compile(r",(?![^()\[\]]*[)\]])")


def _split(type_str):
    if not type_str:
        return []
    parts = _SPLIT_ARGS_RE.split(type_str)

    result = []
    current_part = parts.pop(0)
    while parts:
        if len(re.findall(r"\[", current_part)) > len(re.findall("]", current_part)):
            current_part += "," + parts.pop(0)
        else:
            result.append(current_part)
            current_part = parts.pop(0)
    result.append(current_part)
    return result


def _parse_return_type(return_types):
    # return type could be tuple as well
    # we would like to split this definition into "separate" types
    list_type = get_inline_Tuple_params(return_types)
    if list_type is not None:
        return list_type

    tuple_types = get_Tuple_params(return_types)
    if tuple_types is not None:
        return tuple_types
    return return_types


def _parse_python_2_type_hint_line(typehint_string):
    # type: (str) -> (tuple, type)
    #
    # we remove all whitespaces from the string
    # it simplifies our parsing

    typehint_string = typehint_string.replace(" ", "")
    search_results = TYPE_HINT_RE.search(typehint_string)
    if not search_results:
        raise InvalidTypeHint(
            "%s does not match type hint spec regex %s"
            % (typehint_string, TYPE_HINT_RE)
        )

    arg_types = _split(search_results.group(1))
    return_types = _parse_return_type(search_results.group(2))
    return arg_types, return_types


def parse_arg_types_for_callable(func):
    # type:(callable)->tuple
    """

    :param func:
    :return: list of parameter types if successfully parsed, else None
    """

    # TODO: make this compatible with python 3 type hints
    # python 2.7 type hint
    try:
        source_lines = inspect.getsource(func).split("\n")
    except OSError:
        # It's not possible to get source code when running script as a string passed to IPython.
        # This is how Databricks passes pyspark scripts wo we have to workaround it.
        from dbnd._core.log import dbnd_log_info_error

        dbnd_log_info_error("Failed to parse doc string for function %s" % func)
        return None, None

    def_statements = 0
    for source_line in source_lines:
        try:
            arg_types_tuple, return_type = _parse_python_2_type_hint_line(source_line)
            return arg_types_tuple, return_type
        except InvalidTypeHint:
            if source_line.strip().startswith("def "):
                def_statements += 1
            if def_statements > 1:
                break

    return None, None


def get_doc_annotaions(f, arg_names):
    # type: (Callable, Iterable[str]) -> Optional[Dict[str,type ]]
    arg_types_tuple, return_type = parse_arg_types_for_callable(f)
    if arg_types_tuple or return_type:
        types = dict(zip(arg_names, arg_types_tuple))
        types["return"] = return_type
        return types
    return None



# © Copyright Databand.ai, an IBM Company 2022

import typing

from collections import OrderedDict
from typing import Optional, Set

import six

from dbnd._core.configuration.config_value import ConfigValue


# default type of section/values container
_config_dict_type = OrderedDict
_config_id = 0

# MARKERS:
CONFIG_REPLACE_SECTION_MARKER = "__config_replace_section"


def replace_section_with(dict_value):
    """Replace config section with passed values."""
    dict_value[CONFIG_REPLACE_SECTION_MARKER] = True
    return dict_value


def _lower_config_name(v):
    return v.lower()


# simple class that represents two level dict
class _ConfigStore(OrderedDict):
    def __init__(self, *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)
        self.source = None

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key.lower())
        except KeyError:
            self[key] = section = _config_dict_type()
            return section

    def get_config_value(self, section, key):
        # type: (str, str)->Optional[ConfigValue]
        section = self.get(_lower_config_name(section))
        if not section:
            return None
        return section.get(_lower_config_name(key))

    def set_config_value(self, section, key, value):
        self[_lower_config_name(section)][_lower_config_name(key)] = value

    def update(self, config_values):
        # type: (_ConfigStore) -> _ConfigStore
        for section, section_values in six.iteritems(config_values):
            # shallow copy of configuration
            section = _lower_config_name(section)
            replace_current_section = section_values.pop(
                CONFIG_REPLACE_SECTION_MARKER, None
            )

            if replace_current_section:
                # we can not just reuse section from right -> it's mutable objects
                self[section] = current_section = _config_dict_type()
            else:
                current_section = self[section]

            for key, value in six.iteritems(section_values):
                key = _lower_config_name(key)
                old_value = current_section.get(key)
                if old_value:
                    if old_value.priority > value.priority:
                        continue
                current_section[key] = value
        return self

    def as_value_dict(self, sections=None):
        # type: (Set[str])-> _config_dict_type
        return _config_dict_type(
            (
                section,
                _config_dict_type(
                    (key, str(config_value.value))
                    for key, config_value in six.iteritems(values)
                ),
            )
            for section, values in six.iteritems(self)
            if not sections or section in sections
        )


if typing.TYPE_CHECKING:
    # helping typehints system, our ordered dict is typed dict
    # not serializable in python < 3.7 (python bug)
    _TConfigStore = typing.Union[
        OrderedDict[str, OrderedDict[str, ConfigValue]], _ConfigStore
    ]


def merge_config_stores(config_left, config_right):
    # type : (_ConfigStore, _ConfigStore)-> _ConfigStore
    # known issue: non consisten copy on write
    # for example  c1.merge()  -> return c2.   c1.update() will affect c2 as well

    if not config_right:
        return config_left
    if not config_left:
        return config_right

    new_config = config_left.copy()
    # we copy only section that are changed
    for section in config_right.keys():
        new_config[section] = config_left[section].copy()

    # return original if no changes,
    # copy on write
    new_config.update(config_right)
    return new_config



# © Copyright Databand.ai, an IBM Company 2022

from collections import ChainMap
from typing import Dict, List, Optional, Tuple, Union

import attr
import six

from dbnd._core.constants import MetricSource
from dbnd._core.tracking.log_data_request import LogDataRequest
from dbnd._core.tracking.schemas.column_stats import (
    ColumnStatsArgs,
    get_column_stats_by_col_name,
)
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.utils.timezone import utcnow
from targets.data_schema import DataSchemaArgs, load_data_schema


# keep it below VALUE_PREVIEW_MAX_LEN at web
_DEFAULT_VALUE_PREVIEW_MAX_LEN = 10000


@attr.s(slots=True)
class ValueMeta(object):
    value_preview = attr.ib()  # type: str
    data_dimensions = attr.ib(
        default=None
    )  # type: Optional[Tuple[Optional[int], Optional[int]]]
    query = attr.ib(default=None)  # type: Optional[str]
    data_schema = attr.ib(
        default=None, converter=load_data_schema
    )  # type: Optional[DataSchemaArgs]
    data_hash = attr.ib(default=None)  # type: Optional[str]
    columns_stats = attr.ib(default=attr.Factory(list))  # type: List[ColumnStatsArgs]
    histograms = attr.ib(default=None)  # type: Optional[Dict[str, Tuple]]
    histogram_system_metrics = attr.ib(default=None)  # type: Optional[Dict]
    op_source = attr.ib(default=None)  # type: Optional[str]

    def build_metrics_for_key(self, key, meta_conf=None):
        # type: (str, Optional[ValueMetaConf]) -> Dict[str, List[Metric]]
        ts = utcnow()
        dataframe_metric_value = {}
        data_metrics, hist_metrics = [], []
        metric_source = MetricSource.user
        if self.data_dimensions:
            dataframe_metric_value["data_dimensions"] = self.data_dimensions
            for dim, size in enumerate(self.data_dimensions):
                key_name = f"{key}.shape{dim}"
                self.append_metric(data_metrics, key_name, metric_source, size, ts)
                name = "rows" if dim == 0 else "columns"
                key_name = f"{key}.{name}"
                self.append_metric(data_metrics, key_name, metric_source, size, ts)

        if meta_conf and meta_conf.log_schema:
            data_schema = self.data_schema.as_dict() if self.data_schema else None
            dataframe_metric_value["schema"] = data_schema
            key_name = f"{key}.schema"
            self.append_metric(
                data_metrics, key_name, metric_source, None, ts, data_schema
            )

        if meta_conf and meta_conf.log_preview:
            dataframe_metric_value["value_preview"] = self.value_preview
            dataframe_metric_value["type"] = "dataframe_metric"
            key_name = str(key)
            self.append_metric(
                data_metrics, key_name, metric_source, None, ts, dataframe_metric_value
            )

        metric_source = MetricSource.histograms
        if self.histogram_system_metrics:
            key_name = f"{key}.histogram_system_metrics"
            self.append_metric(
                hist_metrics,
                key_name,
                metric_source,
                None,
                ts,
                self.histogram_system_metrics,
            )

        if self.histograms:
            key_name = f"{key}.histograms"
            self.append_metric(
                hist_metrics, key_name, metric_source, None, ts, self.histograms
            )

        if self.columns_stats:
            # We dump_op_column_stats to old stats_dict for backward compatibility support
            stats_dict = self.get_stats_dict_from_columns_stats()
            key_name = f"{key}.stats"
            self.append_metric(
                hist_metrics, key_name, metric_source, None, ts, stats_dict
            )
            for col_name, stats in stats_dict.items():
                for stat, value in stats.items():
                    key_name = f"{key}.{col_name}.{stat}"
                    self.append_metric(hist_metrics, key_name, metric_source, value, ts)
        return {"user": data_metrics, "histograms": hist_metrics}

    def get_column_stats_by_col_name(
        self, column_name: str
    ) -> Optional[ColumnStatsArgs]:
        return get_column_stats_by_col_name(self.columns_stats, column_name)

    def get_stats_dict_from_columns_stats(self) -> Optional[dict]:
        # Returns legacy stats dict for backward compatability support
        return dict(
            ChainMap(
                *[
                    col_stats.dump_to_legacy_stats_dict()
                    for col_stats in self.columns_stats
                ]
            )
        )

    def append_metric(
        self, data_metrics, key_name, metric_source, value, ts, value_json=None
    ):
        data_metrics.append(
            Metric(
                key=key_name,
                value=value,
                source=metric_source,
                timestamp=ts,
                value_json=value_json,
            )
        )


@attr.s
class ValueMetaConf(object):
    log_preview = attr.ib(default=None)  # type: Optional[bool]
    log_preview_size = attr.ib(default=None)  # type: Optional[int]
    log_schema = attr.ib(default=None)  # type: Optional[bool]
    log_size = attr.ib(default=None)  # type: Optional[bool]
    log_histograms = attr.ib(
        default=None, converter=LogDataRequest.from_user_param
    )  # type: Optional[Union[LogDataRequest, bool]]
    log_stats = attr.ib(
        default=None, converter=LogDataRequest.from_user_param
    )  # type: Optional[Union[LogDataRequest, bool]]

    def get_preview_size(self):
        return self.log_preview_size or _DEFAULT_VALUE_PREVIEW_MAX_LEN

    @classmethod
    def enabled(cls):
        return ValueMetaConf(
            log_size=True,
            log_preview=True,
            log_schema=True,
            log_stats=True,
            log_histograms=True,
        )

    @classmethod
    def disabled_expensive(cls):
        """
        Disabling any config that can be expensive to calculate
        """
        return ValueMetaConf(
            log_preview=False, log_histograms=False, log_stats=False, log_size=False
        )

    def merge_if_none(self, other):
        # type: (ValueMetaConf, ValueMetaConf) -> ValueMetaConf
        """
        Merging the current meta config with the `other` by merging if None strategy.
        which means - take they value from `other` only if the value in self is none.
        """
        if not isinstance(other, ValueMetaConf):
            raise ValueError(
                "Expected ValueMetaConf got instead {other_type}".format(
                    other_type=type(other)
                )
            )

        # collecting all the values from `other`, only if they are None in `self`
        to_merge = {
            key: value
            for key, value in six.iteritems(other.__dict__)
            if self.__dict__[key] is None
        }
        return attr.evolve(self, **to_merge)
